package data;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import listeners.IDataSetListener;

import java.util.Vector;
import java.util.Random;


/** Class implementing the data representation.
 */
public class Data {

  private Vector<DataDimension> m_pDataDimensions;
  private int m_iEntryNum;

  /** DataSetListener **/
  private Vector<IDataSetListener> m_pDataSetListeners;

  //////////////////////////////////////////
  // CONSTRUCTOR                          //
  //////////////////////////////////////////

  /** Creates an empty data object.
   */
  public Data() {

    //1. set default values for the members
    m_pDataDimensions = new Vector<DataDimension>();
    m_iEntryNum = 0;
    m_pDataSetListeners = new Vector<IDataSetListener>();


  }

  //////////////////////////////////////////
  // GET METHODS                          //
  //////////////////////////////////////////

  /*//returns a given datadimension
  public DataDimension getDataDimension(int dim) {
    if (dim < 0) return null;
    if (dim >= m_pDataDimensions.size()) return null;

    return (DataDimension)m_pDataDimensions.get(dim);
  }*/


  /** Returns the number of dimensions in the data.
   *
   * @return Number of dimensions.
   */
  public int getDimNumber() {
    return m_pDataDimensions.size();
  }


  /** Returns the minimum the given dimension.
   *
   * @param iDimId Integer specifying the ID of the dimension. Range: 0 to (dimension number)-1
   *
   * @return Minimum value of the specified dimension.
   */
  public float getDimensionMinimum(int iDimId) {
    //1. make some checks
    if (iDimId < 0) return Float.NEGATIVE_INFINITY;
    if (iDimId >= m_pDataDimensions.size()) return Float.NEGATIVE_INFINITY;

    //2. return the dimension minimum
    DataDimension pDimension = m_pDataDimensions.get(iDimId);
    return pDimension.getMinimumValue();

  }

  /** Returns the maximum the given dimension.
   *
   * @param iDimId Integer specifying the ID of the dimension. Range: 0 to (dimension number)-1
   *
   * @return Maximum value of the specified dimension.
   */
  public float getDimensionMaximum(int iDimId) {
    //1. make some checks
    if (iDimId < 0) return Float.POSITIVE_INFINITY;
    if (iDimId >= m_pDataDimensions.size()) return Float.POSITIVE_INFINITY;

    //2. return the dimension maximum
    DataDimension pDimension = m_pDataDimensions.get(iDimId);
    return pDimension.getMaximumValue();

  }


  /** Returns the name of the given dimension.
   *
   * @param iDimId Integer specifying the ID of the dimension. Range: 0 to (dimension number)-1
   *
   * @return Name of the specified dimension as String.
   */
  public String getDimensionName(int iDimId) {
    //1. make some checks  
    if (iDimId < 0) return null;
    if (iDimId >= m_pDataDimensions.size()) return null;

    //2. return the dimension name
    DataDimension pDimension = m_pDataDimensions.get(iDimId);
    return pDimension.getName();

  }

  /** Returns the names of the dimensions in the data as string array.
   *
   * @return Names of the dimensions.
   */
  public String[] getDimensionNames() {

    //1. create and fill the dimension name array
    int iDimNum = m_pDataDimensions.size();
    String[] pDimNames = new String[iDimNum];

    for (int i = 0; i < iDimNum; i++) {
      DataDimension pDimension = (DataDimension)m_pDataDimensions.get(i);
      pDimNames[i] = pDimension.getName();
    }

    return pDimNames;

  }



  /** Returns the data item values in the given dimension.
   *
   * @param iDimId Integer specifying the ID of the dimension. Range: 0 to (dimension number)-1
   *
   * @return The values of the data items in the specified dimension as float array.
   */
  public float[] getDimensionValues(int iDimId) {
    //1. make some checks  
    if (iDimId < 0) return null;
    if (iDimId >= m_pDataDimensions.size()) return null;

    //2. return the dimension name
    DataDimension pDimension = (DataDimension)m_pDataDimensions.get(iDimId);
    return pDimension.getValues();
  }


  /** Returns the data item values in the given dimension.
   *
   * @param dimname String specifying the name of the dimension.
   *
   * @return The values of the data items in the specified dimension as float array.
   */
  public float[] getDimensionValues(String pDimName) {
    //1. make some checks
    if (pDimName == null) return null;
    if (pDimName.trim().equals("")) return null;
    pDimName = pDimName.trim();

    //2. find the dimension id
    String[] pDimNames = getDimensionNames();
    int iDimNum = pDimNames.length;
    int iDimId = -1;

    for (int i = 0; i < iDimNum; i++) {
      if (pDimNames[i].trim().equals(pDimName)) {
        iDimId = i;
        break;
      }
    }
    
    //3. return the dimension values
    return getDimensionValues(iDimId);
  }




  /** Returns the number of data items (= entries) in the data.
   *
   * @return Number of data items.
   */
  public int getEntryNum() {
    DataDimension pDimension = m_pDataDimensions.get(0);
    return pDimension.getEntryNum();
  }




  /** Returns a summary of a specified data item as String.
   *
   * @param iItemId Integer specifying the ID of the data item. Range: 0 to (entry number)-1
   *
   * @return String representation of the data item.
   */
  public String itemToCSVString(int iItemId, String pSeparator) {
    //1. make some checks
    if (pSeparator == null) pSeparator = "";

    //2. create the csv line representation
    String pReturnValue = "";
    int iDimNum = m_pDataDimensions.size();
    
    for (int i = 0; i < iDimNum; i++) {
      pReturnValue += ((DataDimension)m_pDataDimensions.get(i)).getStringValue(iItemId);

      if (i < iDimNum-1) {
        pReturnValue += pSeparator;
      }
    }

    return pReturnValue;

  }

  /** Adds a new Dimension to the data.
   *
   * @param dim DataDimension object that represents a dimension.
   *
   */
  public void addDimension(DataDimension pDimension) {
    //1. make a simple check
    if (pDimension == null) return;

    //2. add the dimension
    m_pDataDimensions.add(pDimension);
    if (m_iEntryNum == 0) m_iEntryNum = pDimension.getEntryNum();
    dimensionNumberChanged(pDimension.getName());

  }

  //////////////////////////////////////////
  // SET METHODS                          //
  //////////////////////////////////////////

  /** Sets new values to the specified dimension.
   *
   * @param iDimId Integer specifying the ID of the dimension. Range: 0 to (dimension number)-1
   * @param pValues float array holding a value for each data item.
   *
   */
  public void setDimensionValues(int iDimId, float[] pValues) {
    //1. make some checks
    if (iDimId < 0) return;
    if (iDimId >= m_pDataDimensions.size()) return;

    
    //2. do the job
    DataDimension pDimension = m_pDataDimensions.get(iDimId);
    pDimension.setValues(pValues);
    dimensionValuesChanged(pDimension.getName());

  }


  /** Sets new values to the specified dimension.
   *
   * @param pDimName String specifying the name of the dimension.
   * @param pValues float array holding a value for each data item.
   *
   */
  public void setDimensionValues(String pDimName, float[] pValues) {
    //1. make some checks
    if (pDimName == null) return;
    if (pDimName.trim().equals("")) return;
    pDimName = pDimName.trim();
   
    //2. find the dimension id
    String[] pDimNames = getDimensionNames();
    int iDimNum = pDimNames.length;
    int iDimId = -1;

    for (int i = 0; i < iDimNum; i++) {
      if (pDimNames[i].trim().equals(pDimName)) {
        iDimId = i;
        break;
      }
    }
    
    //3. do the job
    setDimensionValues(iDimId, pValues);

  }


  //////////////////////////////////////////
  // DATASET LISTENER METHODS             //
  //////////////////////////////////////////
  /** Adds a IDataSetListener object that is notified, if changes in the data occur.
   *
   * @param listener IDataSetListener object that has to do some tasks if the data changes.
   *
   */
  public void addDataSetListener(IDataSetListener pListener) {
    //1. make a check
    if (pListener == null) return;
    
    //2. add the listener
    m_pDataSetListeners.add(pListener);
  }

  
  private void dimensionValuesChanged(String pDimName) {
    int iListenerNum = m_pDataSetListeners.size();
    
    for (int i = 0; i < iListenerNum; i++) {
      IDataSetListener pListener = m_pDataSetListeners.get(i);
      pListener.dimensionValuesChanged(this, pDimName);
    }

  }

  private void dimensionNumberChanged(String pDimName) {
    int iListenerNum = m_pDataSetListeners.size();
    
    for (int i = 0; i < iListenerNum; i++) {
      IDataSetListener pListener = m_pDataSetListeners.get(i);
      pListener.dimensionNumberChanged(this, pDimName);
    }

  }


}