package data;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import listeners.GlobalConstants;
import listeners.ISelectionListener;
import listeners.ISelectionListListener;


import java.util.Vector;


/** The class holding and managing all Selections of all views. This class manages the calculation of the
 *  local and global degree of interest.
 */
public class SelectionList implements GlobalConstants, ISelectionListener, Runnable {

  /** member variables **/
  private Data m_pData;               //data object
  private Vector<Selection> m_pSelectionList;    //vector holding all selections


  // DOI = an array containing a value between 0 and 1 for each data item
  //1 = selected, 0 = not selected

  private float[] m_pGlobalDOI;        //float array holding the global DOI
  private float[][] m_pLocalDOI;       //2d float array holding the local DOI for each view (viewNum x entryNum)

  private int[] m_pSelectionNum;     //holds for each view the number of selections

  private Vector<ISelectionListListener> m_pSelectionListListeners; //vector for the SelectionListListeners


  /*** THREAD ***/
  private Thread  m_pUpdateThread;     //thread doing the DOI calculation and initiating the update of the views
  private int     m_iViewIdforThread;    //parameter needed for starting the thread


  private int m_iFireCounter;

  //////////////////////////////////////////
  // CONSTRUCTOR                          //
  //////////////////////////////////////////

  /** Creates an empty SelectionList object based on the given Data object.
   *
   *  @param data Data object holding the data on which the degree of interest calculations are made.
   */
  public SelectionList(Data pData) {
    m_pData = pData;
    m_pSelectionList = new Vector<Selection>();

    int iEntryNum = m_pData.getEntryNum();
    m_pGlobalDOI = new float[iEntryNum];
    m_pLocalDOI = new float[VIEW_NUM][iEntryNum];

    m_pSelectionNum = new int[VIEW_NUM];
    for (int i = 0; i < VIEW_NUM; i++) {
      m_pSelectionNum[i] = 0;

      for (int j = 0; j < iEntryNum; j++) {
        m_pLocalDOI[i][j] = 1.0f;
      }
    }

    m_pSelectionListListeners = new Vector<ISelectionListListener>();

    m_pUpdateThread = null;

    //TO DO: replace by recalculate global DOI
    recalculateGlobalDOI();

    m_iFireCounter = 0;
  }


  /** Adds a Selection to the SelectionList and initiates a recalculation of the local degree of interest.
   *
   *  @param pSelection Selection that is added to the SelectionList.
   */
  public void addSelection(Selection pSelection) {
    pSelection.addSelectionListener(this);
    m_pSelectionList.add(pSelection);
    m_pSelectionNum[pSelection.getViewId()]++;


    selectionChanged(pSelection);
  }


  /** Deletes a Selection from the SelectionList and initiates a recalculation of the local degree of interest.
   *
   *  @param pSelection Selection that is deleted from the SelectionList.
   *  @param fireChangeEvent boolean that indicates if an update of the DOI should be calculated and the listeners should be notified.
   */
  public void deleteSelection(Selection pSelection, boolean fireChangeEvent) {
    m_pSelectionList.remove(pSelection);

    int iViewId = pSelection.getViewId();
    if (m_pSelectionNum[iViewId] > 0) {
      m_pSelectionNum[iViewId]--;
    }

    if (fireChangeEvent) {
      selectionChanged(pSelection);
    }
  }


  /** Deletes all Selections from the SelectionList that are created in the specified view
   *  and initiates a recalculation of the local degree of interest.
   *
   *  @param viewID ID of a view.
   */
  public void deleteAllLocalSelections(int iViewId) {

    Vector<Selection> pLocalSelections = getLocalSelections(iViewId);
    int iSelectionNum = pLocalSelections.size();
    for (int i = 0; i < iSelectionNum; i++) {
      m_pSelectionList.remove(pLocalSelections.get(i));
    }

    m_pSelectionNum[iViewId] = 0;

    recalculateLocalDOI(iViewId);
  }


  /** Returns all Selections that are created in the specified view.
   *
   *  @param iViewId ID of a view.
   *
   *  @return A Vector holding all Selections created in the specified view.
   */
  public Vector<Selection> getLocalSelections(int iViewId) {
    Vector<Selection> pLocalSelections = new Vector<Selection>();

    int iSelectionNum = m_pSelectionList.size();
    for (int i = 0; i < iSelectionNum; i++) {

      Selection pCurrSel = (Selection)m_pSelectionList.get(i);
      if (pCurrSel.getViewId() == iViewId) {
        pLocalSelections.add(pCurrSel);
      }
    }

    return pLocalSelections;
  }

  /** Empty method.
   *
  */
  public float[] getGlobalDOI() {
    return m_pGlobalDOI;
  }


  /** Returns the local degree of interest for the specified view.
   *
   *  @param iViewId ID of a view.
   *
   *  @return A float array holding a degree of interest value for each data item.
   */
  public float[] getLocalDOI(int iViewId) {
    return m_pLocalDOI[iViewId];
  }


  public void setData(Data pData) {
    m_pData = pData;

    int iEntryNum = m_pData.getEntryNum();
    m_pGlobalDOI = new float[iEntryNum];
    m_pLocalDOI = new float[VIEW_NUM][iEntryNum];

    for (int i = 0; i < VIEW_NUM; i++) {
      for (int j = 0; j < iEntryNum; j++) {
        m_pLocalDOI[i][j] = 1.0f;
      }
    }

    m_pUpdateThread = new Thread(this);
    recalculateParCoordDOI(m_pUpdateThread);
    recalculateScatterplotDOI(m_pUpdateThread);
    recalculateGlobalDOI();

  }




  ////////////////////////////////////////
  //DOI calculations                    //
  ////////////////////////////////////////
  private void recalculateLocalDOI(int iViewId) {

    //set the current thread
    Thread pMyThread = Thread.currentThread();


    //initialize the doi with 1 values
    int iEntryNum = m_pData.getEntryNum();
    for (int i = 0; i < iEntryNum; i++) {
      m_pLocalDOI[iViewId][i] = 1.0f;
    }

    //if there is no selection, the calculation is done
    if (m_pSelectionNum[iViewId] == 0) {
      recalculateGlobalDOI();
      selectionListChanged();
      return;
    }

    boolean bUpdateComplete;
    if (iViewId == VIEW_ID_PAARCOORD) {
      bUpdateComplete = recalculateParCoordDOI(pMyThread);
    }
    else {
      bUpdateComplete = recalculateScatterplotDOI(pMyThread);
    }


    if (bUpdateComplete) {
      recalculateGlobalDOI();
      selectionListChanged();
    }
  }





  private boolean recalculateParCoordDOI(Thread pMyThread) {
    //for every dimension an doi must be calculated
    int iDimNum = m_pData.getDimNumber();
    int iEntryNum = m_pLocalDOI[VIEW_ID_PAARCOORD].length;


    for (int iDim = 0; iDim < iDimNum; iDim++) {

      //check if we should stop calculation
      if (m_pUpdateThread != pMyThread) {
        return false;
      }

      //init the dimension doi with 0 values
      float[] pDimensionDOI = new float[iEntryNum];
      for (int i = 0; i < iEntryNum; i++) {
        pDimensionDOI[i] = 0.0f;
      }

      //indicates if there is a selection on this dimension
      boolean bIsDimSelected = false;

      //check all selections if they are on the current dimension
      int iSelectionNum = m_pSelectionList.size();
      for (int i = 0; i < iSelectionNum; i++) {

        //check if we should stop calculation
        if (m_pUpdateThread != pMyThread) {
          return false;
        }



        //check which kind of selection (only AxisSelections are allowed)
        Selection pCurrSelection = m_pSelectionList.get(i);
        if (!(pCurrSelection instanceof AxisSelection)) continue;
        AxisSelection pCurrAxisSelection = (AxisSelection)pCurrSelection;

        //check if selection is on the current dimension
        if (pCurrAxisSelection.getDimId() != iDim) continue;

        //on the current dimension there is a selection
        bIsDimSelected = true;


        //get the values of the dimension
        float[] pValues = m_pData.getDimensionValues(iDim);
        if (pValues == null) continue;

        //get the doi of the current selection
        float[] pCurrDOI = pCurrAxisSelection.getDOI(pValues);

        //the doi of the current dimensions must be combined in the OR mode
        for (int j = 0; j < iEntryNum; j++) {
          pDimensionDOI[j] = Math.max(pCurrDOI[j], pDimensionDOI[j]);
        }


      }

      //check if we should stop calculation
      if (m_pUpdateThread != pMyThread) {
        return false;
      }


      //if there is a selection on the current dimension the overall doi has changed
      if (bIsDimSelected) {
        float[] pLocalDOI = m_pLocalDOI[VIEW_ID_PAARCOORD];
        for (int i = 0; i < iEntryNum; i++) {
          //the overall doi must be combined with the dimension doi in AND mode
          pLocalDOI[i] = Math.min(pDimensionDOI[i], pLocalDOI[i]);
        }
        m_pLocalDOI[VIEW_ID_PAARCOORD] = pLocalDOI;
      }
    }

    return true;
  }


  private boolean recalculateScatterplotDOI(Thread pMyThread) {
    int iEntryNum = m_pLocalDOI[VIEW_ID_SCATTERPLOT].length;

    //check all selections if they are on the current dimension
    int iSelectionNum = m_pSelectionList.size();
    for (int i = 0; i < iSelectionNum; i++) {

      //check if we should stop calculation
      if (m_pUpdateThread != pMyThread) {
        return false;
      }


      Selection pCurrSelection = m_pSelectionList.get(i);

      //check if correct view and correct selection type
      if (pCurrSelection.getViewId() != VIEW_ID_SCATTERPLOT) continue;
      if (!(pCurrSelection instanceof ScatterplotSelection)) continue;


      ScatterplotSelection pCurrScSelection = (ScatterplotSelection)pCurrSelection;
      float[] pXVals = m_pData.getDimensionValues(pCurrScSelection.m_iXdimID);
      float[] pYVals = m_pData.getDimensionValues(pCurrScSelection.m_iYdimID);
      float[] pDimensionDOI = pCurrScSelection.getDOI(pXVals, pYVals);

      //check if we should stop calculation
      if (m_pUpdateThread != pMyThread) {
        return false;
      }


      float[] pLocalDoi = m_pLocalDOI[VIEW_ID_SCATTERPLOT];
      for (int iEntry = 0; iEntry < iEntryNum; iEntry++) {
        //the overall doi must be combined with the dimension doi in AND mode
        pLocalDoi[iEntry] = Math.min(pDimensionDOI[iEntry], pLocalDoi[iEntry]);
      }
      m_pLocalDOI[VIEW_ID_SCATTERPLOT] = pLocalDoi;
    }

    return true;



  }


  //combines the dois of the two views with AND
  private void recalculateGlobalDOI() {

    int iEntryNum = m_pGlobalDOI.length;

    float[] pLocalDOI1 = m_pLocalDOI[0];
    float[] pLocalDOI2 = m_pLocalDOI[1];
    for (int i = 0; i < iEntryNum; i++) {
      m_pGlobalDOI[i] = Math.min(pLocalDOI1[i], pLocalDOI2[i]);
    }

  }


  //not used method to save a doi channel to the data
  private void saveLocalDOI (int iViewId) {
    if (m_pData.getDimensionValues(DOINAME) == null) {
      DataDimension pDOIDim = new FloatDataDimension(m_pLocalDOI[iViewId], DOINAME);
      m_pData.addDimension(pDOIDim);
    }
    else {
      m_pData.setDimensionValues(DOINAME, m_pLocalDOI[iViewId]);
    }

  }




  //////////////////////////////////////////
  // SELECTIONLIST LISTENER METHODS       //
  //////////////////////////////////////////
  /** Adds a ISelectionListListener object that is notified, if changes in the SelectionList occur.
   *
   * @param listener ISelectionListListener object that has to do some tasks if the SelectionList changes.
   *
   */
  public void addSelectionListListener(ISelectionListListener pListener) {
    m_pSelectionListListeners.add(pListener);
  }

  public void fireSelectionListChanged() {
    selectionListChanged();
  }

  private void selectionListChanged() {

    int iListenerNum = m_pSelectionListListeners.size();
    for (int i = 0; i < iListenerNum; i++) {
      ISelectionListListener pListener = m_pSelectionListListeners.get(i);
      pListener.selectionListChanged(this);
    }

  }

  //////////////////////////////////////////
  // SELECTIONLISTENER METHODS            //
  //////////////////////////////////////////
  /** Invoked when a change in a Selection occurs. Should start a separate Thread that updates the
   *  degree of interest. But this does not work yet!!!!
   *
   *  @param pSelection Selection that has been changed.
   */
  public void selectionChanged(Selection pSelection) {

    m_pUpdateThread = null;
    m_iViewIdforThread = pSelection.getViewId();
    m_pUpdateThread = new Thread(this);
    m_pUpdateThread.start();


  }

  //////////////////////////////////////////
  // RUNNABLE INTERFACE                   //
  //////////////////////////////////////////
  public void run() {
    recalculateLocalDOI(m_iViewIdforThread);
  }


}