package data;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.util.Vector;


/** Class implementing a dimension of the data.
 */
public class StringDataDimension extends DataDimension {

  private String[]  m_pValues;           //data values
  private String    m_pName;
  private Vector<String> m_pUniqueStrings;
  private float[] m_pFloatValues;

  //statistics
  private float     m_fDimMin, m_fDimMax;         //minimum and maximum values
  private int       m_iDimMinPos, m_iDimMaxPos;   //minimum and maximum positions
  private float     m_fDimMean;             //mean value of the dimension


  //////////////////////////////////////////
  // CONSTRUCTORS                         //
  //////////////////////////////////////////

  

  /** Creates a data dimension with String values. Each unique String
   *  value is mapped to an integer value. So plotting of String representations is possible.
   *
   *  @param values String array holding the values of the dimension.
   *  @param name String representing the name of the dimension.
   */
  public StringDataDimension(Vector<String> pValues, String pName) {

    //1. make some checks
    if (pValues == null) {
      pValues = new Vector<String>();
      pValues.add("");
    }
    
    //2. set the members
    int iEntryNum = pValues.size();
    m_pValues = new String[iEntryNum];
    m_pValues = pValues.toArray(m_pValues);
    
    m_pName = pName;
    
    //3. do some calculations
    //3.1. init all data structures and variables
    m_pUniqueStrings = new Vector<String>();
    m_pFloatValues = new float[iEntryNum];
    
    m_fDimMin = 0.0f; //is set and no longer touched
    m_fDimMax = Float.NEGATIVE_INFINITY;
    m_iDimMinPos = 0; //is set and no longer touched
    m_iDimMaxPos = 0;

    float fSum = 0.0f;
    String pCurrValue = "";
    float fFloatValue = 0.0f;


    //3.2. fill the float array with values (every different stringvalue
    //     gets a unique integer value)
    for (int i = 0; i < iEntryNum; i++) {
      
      //calculate the float value and set the max value and position if necessary
      pCurrValue = m_pValues[i];
      int iIndex = m_pUniqueStrings.indexOf(pCurrValue);
      
      fFloatValue = (float)iIndex;
      if (iIndex < 0) {
        fFloatValue = m_pUniqueStrings.size();
        m_pUniqueStrings.add(pCurrValue);
        
        m_fDimMax = fFloatValue;
        m_iDimMaxPos = i;
      }
      
      //set the float value in the array and add the value to the sum
      m_pFloatValues[i] = fFloatValue;
      fSum += fFloatValue;
      
    }
    
    m_fDimMean = fSum/(float)iEntryNum;


    //3. set the statistics data in the abstract class
    setDimensionInfo(m_pName, iEntryNum, m_fDimMin, m_fDimMax, m_iDimMinPos, m_iDimMaxPos, m_fDimMean);


  }

  
  /////////////////////////////////////
  // IMPLEMENTATION OF THE INTERFACE //
  /////////////////////////////////////
  //empty at the moment
  public void setValues(float[] pValues) {}
  
  /** Returns the values of the dimension as float arry.
   *
   * @return float array holding a value for each data item.
   */
  public float[] getValues() {
    return m_pFloatValues;
  }
  
  

  /** Returns the String representation of the value of the specified data item in this dimension.
   *
   * @return String representation of the value of the specified data item in this dimension.
   */
  public String getStringValue(int iEntryId) {
    if (iEntryId < 0) return null;
    if (iEntryId >= m_pValues.length) return null;

    return m_pValues[iEntryId];

  }







}