package tools;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


public class ChiSquaredDistribution {


  private static final double DIFF = 0.0000000001;


  //returns the pdf (propability density function) values of the chi - squared distribution 
  //with a given degree of freedom for given x values
  public static double[] getDensityValues(int iDegreeOfFreedom, double[] pXValues) {

    //1. make some checks
    //to do: check for valid values of the degree of freedom
    if (pXValues == null) return null;
    int iValueNum = pXValues.length;
    
    
    //2. create the density value array and fill it
    double[] pDensity = new double[iValueNum];

    double dNormFactor = 1;
    dNormFactor /= Math.pow(2, (double)iDegreeOfFreedom/2) * Math.exp(gammln((double)iDegreeOfFreedom/2));

    double dFirstExponent = (double)iDegreeOfFreedom/2 - 1;


    for (int i = 0; i < iValueNum; i++) {
      double dCurrValue = pXValues[i];
      
      if (dCurrValue < 0) {
        pDensity[i] = 0;
        continue;
      }

      pDensity[i] = dNormFactor * Math.pow(dCurrValue, dFirstExponent) * Math.exp(-dCurrValue/2);

    }

    return pDensity;


  }


  //numerical recipes method returning the log naturalis of the gamma function for a given value
  //needed by getDensity
  private static double gammln(double dAlpha) {

    //taken from numerical recipes: function gammaln
    double x,y,tmp,ser;
    double[] cof={76.18009172947146,-86.50532032941677,
    24.01409824083091,-1.231739572450155,
    0.1208650973866179e-2,-0.5395239384953e-5};
    int j;
    y=x=dAlpha;
    tmp=x+5.5;
    tmp -= (x+0.5)*Math.log(tmp);
    ser=1.000000000190015;
    for (j=0;j<=5;j++) ser += cof[j]/++y;


    return -tmp+Math.log(2.5066282746310005*ser/x);



  }




  //returns the cdf (conditional distribution function) values of the 
  //chi - squared distribution with a given degree of freedom for given x values
  public static double[] getDistributionValues(int iDegreeOfFreedom, double[] pXValues) {

    //1. make some checks
    //to do: check for valid values of the degree of freedom
    if (pXValues == null) return null;
    int iValueNum = pXValues.length;
  
  
    //2. fill the values
    double[] pDistribution = new double[iValueNum];

    for (int i = 0; i < iValueNum; i++) {
      double dCurrValue = pXValues[i];
      if (dCurrValue < 0) {
        pDistribution[i] = 0;
        continue;
      }

      pDistribution[i] = 1 - (double)gammq((double)iDegreeOfFreedom/2, (double)dCurrValue/2);

    }

    return pDistribution;


  }


  //returns the cdf values of the chi - squared distribution with a given degree of freedom for given x values
  public static double getDistributionValue(int iDegreeOfFreedom, double dXValue) {

    if (dXValue < 0) {
      return 0;
    }

    return (1 - (double)gammq((double)iDegreeOfFreedom/2, (double)dXValue/2));


  }


  //numerical recipes method
  //needed by getDistributionValues
  private static double gammq(double a, double x) {
    double gamser = 0;
    double gammcf = 0;
    double gln = 0;

    if (x < 0.0 || a <= 0.0) System.out.println("Invalid arguments in routine gammq");
    if (x < (a+1.0)) {
      double[] vals = gser(gamser,a,x,gln);
      gamser = vals[0];
      gln = vals[1];
      return 1.0-gamser;
    } else {
      double[] vals = gcf(gammcf,a,x,gln);
      gammcf = vals[0];
      gln = vals[1];
      return gammcf;
    }
  }


  //numerical recipes method
  //needed by getDistributionValues
  private static double[] gcf(double gammcf, double a, double x, double gln)
  {

    int ITMAX = 100;
    double EPS  = 3.0e-7;
    double FPMIN = 1.0e-30;



    int i;
    double an,b,c,d,del,h;

    gln=gammln(a);
    b=x+1.0-a;
    c=1.0/FPMIN;
    d=1.0/b;
    h=d;
    for (i=1;i<=ITMAX;i++) {
      an = -i*(i-a);
      b += 2.0;
      d=an*d+b;
      if (Math.abs(d) < FPMIN) d=FPMIN;
      c=b+an/c;
      if (Math.abs(c) < FPMIN) c=FPMIN;
      d=1.0/d;
      del=d*c;
      h *= del;
      if (Math.abs(del-1.0) < EPS) break;
    }
    if (i > ITMAX) System.out.println("a too large, ITMAX too small in gcf");
    gammcf=Math.exp(-x+a*Math.log(x)-(gln))*h;


    double[] returnVals = new double[2];
    returnVals[0] = gammcf;
    returnVals[1] = gln;

    return returnVals;

  }



  //numerical recipes method
  //needed by getDistributionValues
  private static double[] gser(double gamser, double a, double x, double gln)
  {

    int ITMAX = 100;
    double EPS  = 3.0e-7;


    int n;
    double sum,del,ap;

    gln=gammln(a);
    if (x <= 0.0) {
      if (x < 0.0) System.out.println("x less than 0 in routine gser");
      gamser=0.0;

      double[] returnVals = new double[2];
      returnVals[0] = gamser;
      returnVals[1] = gln;

      return returnVals;


    } else {
      ap=a;
      del=sum=1.0/a;
      for (n=1;n<=ITMAX;n++) {
        ++ap;
        del *= x/ap;
        sum += del;
        if (Math.abs(del) < Math.abs(sum)*EPS) {
          gamser=sum*Math.exp(-x+a*Math.log(x)-(gln));
          double[] returnVals = new double[2];
          returnVals[0] = gamser;
          returnVals[1] = gln;

          return returnVals;
        }
      }
      System.out.println("a too large, ITMAX too small in routine gser");
      return null;
    }
  }



  public static double[] getQuantileValues(int iDegreeOfFreedom, double[] pQuantiles) {

    //1. make some checks
    //to do: check for valid values of the degree of freedom
    if (pQuantiles == null) return null;
    int iValueNum = pQuantiles.length;
  
  
    //2. calculate the x values
    double[] pXValues = new double[iValueNum];

    for (int i = 0; i < iValueNum; i++) {
      pXValues[i] = getQuantileValue(iDegreeOfFreedom, pQuantiles[i]);

    }

    return pXValues;


  }



  public static double getQuantileValue(int iDegreeOfFreedom, double dQuantile) {


    //1. return extreme values
    if (dQuantile <= 0) return 0;
    if (dQuantile >= 1) return Double.POSITIVE_INFINITY;



    //2. initialize iterative check
    double dUpperX = 30;
    double dLowerX = 0;
    double dUpperAppr, dLowerAppr;
    double dUpperDiff, dLowerDiff, dMinDiff;
    dLowerAppr = 0;


    dUpperAppr = getDistributionValue(iDegreeOfFreedom, dUpperX);
    //System.out.println(dUpperAppr + "");
    while (dUpperAppr < dQuantile) {
      dLowerAppr = dUpperAppr;
      dLowerX = dUpperX;

      dUpperX = dLowerX+3;
      dUpperAppr = getDistributionValue(iDegreeOfFreedom, dUpperX);
      //System.out.println(dUpperAppr + "");
    }
    //System.out.println(dLowerAppr + ", " + dUpperAppr);



    dUpperDiff = dUpperAppr - dQuantile;
    dLowerDiff = dQuantile - dLowerAppr;
    dMinDiff = Math.min(dUpperDiff, dLowerDiff);


    //start iterative check
    while(dMinDiff > DIFF) {
      //System.out.println("Min diff: " + dMinDiff);
      double dNewX = (dUpperX + dLowerX)/2;
      double dNewAppr = getDistributionValue(iDegreeOfFreedom, dNewX);

      if (dNewAppr < dQuantile) {
        dLowerAppr = dNewAppr;
        dLowerX = dNewX;
        dLowerDiff = dQuantile - dLowerAppr;

      }
      else {
        dUpperAppr = dNewAppr;
        dUpperX = dNewX;
        dUpperDiff = dUpperAppr - dQuantile;
      }

      dMinDiff = Math.min(dUpperDiff, dLowerDiff);

    }

    //return quantile value
    if (dUpperDiff < dLowerDiff) {
      return dUpperX;
    }
    else {
      return dLowerX;
    }


  }



}