package tools;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import java.util.*;



/** Class implementing the classic and robust calculation of the covariance matrix of a subset of
 *  the dataset or the whole dataset.
 */

public class CovarianceCalculator {


  private float[] m_fMeanVector;      //stores the latest calculated mean values


  //////////////////////////////////////////
  // CONSTRUCTOR                          //
  //////////////////////////////////////////
  /** Creates a CovarianceCalculator object.
   *
   */
  public CovarianceCalculator() {
    m_fMeanVector = null;
  }

  /** Returns the arithmetic mean vector that was used for the calculation of the previously returned covariance matrix.
   *
   *  @return float array holding the arithmetic mean for each dimension.
   */
  public float[] getMeanVector() {
    return m_fMeanVector;
  }



  /** Returns the classic covariance matrix for the given data.
   *
   *  @param pData a <code>dimNumber</code> x <code>entryNumber</code> float array holding holding the values
   *  of the data items in the relevant dimensions.
   *
   *  @return <code>dimNumber</code> x <code>dimNumber</code> float array representing the classic covariance matrix.
   */
  public float[][] classicCovarianceMatrix(float[][] pData) {

    //1. make a simple check
    if (pData == null) return null;
  
    //2. create empty covariance matrix
    int iDimNum = pData.length;
    float[][] pCovarianceMatrix = new float[iDimNum][iDimNum];

    for (int iRow = 0; iRow < iDimNum; iRow++) {
      for (int iCol = 0; iCol < iDimNum; iCol++) {
        pCovarianceMatrix[iRow][iCol] = 0.0f;
      }
    }


    //3. create an array for the means of the dimensions
    m_fMeanVector = new float[iDimNum];
    for (int iDim = 0; iDim < iDimNum; iDim++) {
      m_fMeanVector[iDim] = calculateMean(pData[iDim]);
    }

    //4. fill the covariance matrix
    int iEntryNum = pData[0].length;

    for (int i = 0; i < iEntryNum; i++) {

      //create the mean normalized entry
      float[] fXi = new float[iDimNum];
      for (int iDim = 0; iDim < iDimNum; iDim++) {
        fXi[iDim] = pData[iDim][i] - m_fMeanVector[iDim];
      }


      //add to each entry in covariance matrix
      for (int iRow = 0; iRow < iDimNum; iRow++) {
        float fXiRow = fXi[iRow];
        for (int iCol = 0; iCol < iDimNum;iCol++) {
          pCovarianceMatrix[iRow][iCol] += fXiRow * fXi[iCol];
        }
      }

    }


    //5. divide by N-1
    float fDivisor = (float)iEntryNum-1;
    for (int iRow = 0; iRow < iDimNum; iRow++) {
      for (int iCol = 0; iCol < iDimNum; iCol++) {
        pCovarianceMatrix[iRow][iCol] /= fDivisor;
      }
    }


    return pCovarianceMatrix;


  }


  // calculates the arithmetic mean of the given float array
  private float calculateMean(float[] pData) {
    //1. make a check
    if (pData == null) return 0.0f;
    
    //2. create the mean
    int iLength = pData.length;

    float fSum = 0.0f;
    for (int i = 0; i < iLength; i++) {
      fSum += pData[i];
    }
    return fSum/(float)iLength;

  }




  //approximates the MCD covariance matrix

  //choose 500 arbitrary subsets of the data of size fittedNum
  //make with each of those 500 subsets 2 iteration steps
  //keep the 10 bestsolutions
  //make for each of those 10 solutions iteration steps until the iteration converges
  //the best solution is the appoximation of the MCD covariance matrix

  /** Returns the robust covariance matrix for the given data, calculated according
   *  "A Fast Algorithm for the Minimum Covariance Determinant Estimator" by Peter J. Rousseeuw and Katrien Van Driessen.
   *
   *  @param fRobustness float value between 0.5 and 1 indicating on which percentage of data items should be used for
   *  the robust estimate of the covariance matrix.
   *  @param data a <code>dimNumber</code> x <code>entryNumber</code> float array holding holding the values
   *  of the data items in the relevant dimensions.
   *
   *  @return <code>dimNumber</code> x <code>dimNumber</code> float array representing the classic covariance matrix.
   */
  public float[][] robustCovarianceMatrix(float fRobustness, float[][] pData) {
    //1. make a simple check
    if (pData == null) return null;
    return classicCovarianceMatrix(pData);
    
    //TODO: bring this back into the program with a new math library
    /*

    //2. check parameter h
    if (fRobustness < 0.5) fRobustness = 0.5f;
    if (fRobustness > 1) fRobustness = 1.0f;


    //3. parameters for the sampling
    int iMaxSubsets = 500;
    int iIterMaxSubsets = 2;
    int iBestSolutionNumber = 10;


    //4. dimension and entry number
    int iDimNum = pData.length;
    int iEntryNum = pData[0].length;

    //5. number of data points on which the covariance matrix is fitted
    int iFittedNum = (int)(iEntryNum * fRobustness) + 1;
    if (iFittedNum >= iEntryNum) return classicCovarianceMatrix(pData);


    //6. create vector which is used as random number generator
    Vector<Integer> pRandomNumbers = new Vector<Integer>();
    for (int i = 0; i < iEntryNum; i++) {
      pRandomNumbers.add(new Integer(i));
    }



    //float mindet = Float.POSITIVE_INFINITY;
    //float[][] pBestMatrix = null;


    //7. data array with the entries for which the classic covariance matrix is calculated
    //   for each of those covariance matrix it is tested if it is part of the best solutions
    float[][] pFittedData = new float[iDimNum][iFittedNum];

    float[] pBestDeterminants = new float[iBestSolutionNumber];
    Matrix[] pBestMatrices = new Matrix[iBestSolutionNumber];

    for (int i = 0; i < iMaxSubsets; i++) {

      //7.1. shuffle the randomNumbers to create a new ordering
      Collections.shuffle(pRandomNumbers);

      //7.2. fill the fittedData with the first fittedNum IDs of the randomNumber vector
      for (int iEntry = 0; iEntry < iFittedNum; iEntry++) {
        int iEntryID = (pRandomNumbers.get(iEntry)).intValue();

        for (int iDim = 0; iDim < iDimNum; iDim++) {
          pFittedData[iDim][iEntry] = pData[iDim][iEntryID];
        }
      }


      //7.3. get classic covariance matrix
      float[][] pCovarianceMatrix = classicCovarianceMatrix(pFittedData);


      //7.4. improve the solution of the matrix
      pCovarianceMatrix = improveSolution(pCovarianceMatrix, pData, iFittedNum, iIterMaxSubsets);





      //7.5. add the matrix to the 10 best solutions
      if (i < iBestSolutionNumber) {
        pBestMatrices[i] = new Matrix(pCovarianceMatrix);
        pBestDeterminants[i] = (float)Math.log(pBestMatrices[i].det());
      }
      else {
        Matrix pCurrMatrix = new Matrix(pCovarianceMatrix);
        float fCurrDet = (float)Math.log(pCurrMatrix.det());

        float fMaxDet = Float.NEGATIVE_INFINITY;
        int iMaxDetPos = -1;
        for (int k = 0; k < iBestSolutionNumber; k++) {
          if (pBestDeterminants[k] > fMaxDet) {
            fMaxDet = pBestDeterminants[k];
            iMaxDetPos = k;
          }
        }

        if (fCurrDet < fMaxDet) {
          pBestMatrices[iMaxDetPos] = pCurrMatrix;
          pBestDeterminants[iMaxDetPos] = fCurrDet;
        }

      }

    }



    //8. improve the 10 best solutions
    float fBestDet = Float.POSITIVE_INFINITY;
    int iBestDetPos = -1;
    for (int i = 0; i < iBestSolutionNumber; i++) {
      float[][] pCurrMatrix = improveSolution(pBestMatrices[i].getArrayCopyFloat(), pData, iFittedNum, -1);
      pBestMatrices[i] = new Matrix(pCurrMatrix);
      pBestDeterminants[i] = (float)Math.log(pBestMatrices[i].det());

      if (pBestDeterminants[i] < fBestDet) {
        iBestDetPos = i;
        fBestDet = pBestDeterminants[i];
      }

    }


    System.out.println("Beste loagrithmierte Determinante: " + fBestDet);
    return pBestMatrices[iBestDetPos].getArrayCopyFloat();
    */




  }



  private float[][] improveSolution(
    float[][] pCovarianceMatrix, 
    float[][] pData, 
    int iSmallestDistanceEntryNum, 
    int iSteps) 
  {

    //1. make some checks
    if (pCovarianceMatrix == null) return null;
    if (pData == null) return pCovarianceMatrix;
    
    return pCovarianceMatrix;
    
    //TODO: bring this back into the program with a new math library
    /*
    //2. make the basic settings for the improvement loop
    boolean bRun = true;

    float[][] pBestMatrix = pCovarianceMatrix;
    Matrix pBestMatrixObject = new Matrix(pBestMatrix);
    float fMinDet = (float)pBestMatrixObject.det();

    float[][] pFittedData = null;
    int iCounter = 0;

    
    //3. iterative improve the solution
    while (bRun) {

      //3.1. get the data points with the smallest mahalanobis distance for the current solution
      pFittedData = getSmallestDistancePoints(pBestMatrix, pData, iSmallestDistanceEntryNum);

      //3.2. get classic covariance matrix for those points
      float[][] pCovarianceMatrix2 = classicCovarianceMatrix(pFittedData);

      //3.3. calculate determinant of this classic covariance matrix
      Matrix pMatrix2 = new Matrix(pCovarianceMatrix2);
      float fDet2 = (float)pMatrix2.det();

      //3.4. if this was no improvement -> stop the run, else set the improved solution and go on
      if (fDet2 >= fMinDet) {
        bRun = false;
      }
      else {
        pBestMatrix = pCovarianceMatrix2;
        fMinDet = fDet2;
      }

      //3.5. check if the maximum iteration num has been reached. 
      //    (isteps == -1 -> there is no maximum iteration depth)
      iCounter++;

      if ((iSteps > 0) && (iCounter >= iSteps)) {
        bRun = false;
      }

      //System.out.println("Log(det) = " + Math.log(fMinDet));

    }


    return pBestMatrix;
    */

  }






  //calculates the mahalanobis distance for each datapoint
  //returns the index of the n pData points that have the smallest mahalanobis distance
  private float[][] getSmallestDistancePoints(float[][] pCovarianceMatrix, float[][] pData, int iDataPointNum) {
    //1. make some checks
    if (pCovarianceMatrix == null) return null;
    if (pData == null) return pCovarianceMatrix;



    //2. calculate mahalanobis distance
    Vector<IndexAndValue> pDistances = 
    DistanceCalculator.getMahalanobisDistances(pData, pCovarianceMatrix, m_fMeanVector);
    //TO DO: Is it really valid to use always the same mean vector here?


    //3. sort the distances
    Collections.sort(pDistances);


    //4. take the first n entries and return them
    int iDimNum = pData.length;
    float[][] pSmallestDistancePoints = new float[iDimNum][iDataPointNum];
    for (int i = 0; i < iDataPointNum; i++) {
      int iEntryID = pDistances.get(i).getIndex();

      for (int iDim = 0; iDim < iDimNum; iDim++) {
        pSmallestDistancePoints[iDim][i] = pData[iDim][iEntryID];
      }

    }

    return pSmallestDistancePoints;

  }



}
