
package view;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import listeners.IViewColoring;
import data.Data;

import java.awt.Color;


public class DimensionColoring implements IViewColoring {

  private   int       m_iDimId;    //dimension ID for which the coloring is made

  /*** COLORS ***/
  private   Color       m_pMinColor;    //Color for the minimum value
  private   Color       m_pMaxColor;    //Color for the maximum value


  /*** COLORTRANSITION ***/
  private   ColorTransition m_pColorTransition;   //ColorTransition between minimum and maximum value



  /*** RENDER COLOR  ***/
  private   Color[]     m_pRenderColors;   //array with colors with which the items should be rendered

  /*** FLAGS ***/
  private   boolean     m_bUpdateColoring; //indicating if the coloring must be updated




  public DimensionColoring(int iDimId, Color pMinColor, Color pMaxColor, ColorTransition pColorTransition) {
    //1. make some checks
    if (pMinColor == null) pMinColor = Color.BLACK;
    if (pMaxColor == null) pMaxColor = Color.RED;
    
    //2. set the members
    m_iDimId = iDimId;

    m_pMinColor = pMinColor;
    m_pMaxColor = pMaxColor;
    m_pColorTransition = pColorTransition;

    
    m_pRenderColors = null;
    m_bUpdateColoring = true;


  }


  //////////////////////////////
  // GET METHODS              //
  //////////////////////////////
  /* returns an integer array with the IDs of the data items.
     the first ID in the array should be drawn first and appears in the background.
     the last ID in the array should be drawn as last item and appears in the front.*/
  public int[] getRenderPositions(Data pData) {

    return null;
  }


  /* returns an array of colors for the data items ordered by the render position */
  public Color[] getRenderColors(Data pData) {
    if ((m_bUpdateColoring) && (pData != null)) recalculateMembers(pData);

    return m_pRenderColors;

  }


  // unused method for this type of coloring
  public void updateDOI(float[] pDOI) {}

  public void setColorTransition(ColorTransition pColorTransition) {
    //1. make some checks
    if ((m_pColorTransition == null) && (pColorTransition == null)) return;

    //2. make the settings
    m_pColorTransition = pColorTransition;
    m_bUpdateColoring = true;

  }


  /* set the necessary color properties */
  public void setColoringProperties(Color pNotSelectedColor, Color pSelectedColor, ColorTransition pColorTransition) {
  }


  //////////////////////////////////
  // PRIVATE METHODS              //
  //////////////////////////////////
  
  private void recalculateMembers(Data pData) {
    //1. make a simple check
    if (pData == null) return;
    int iEntryNum = pData.getEntryNum();
  
    //2. fetch the dimension values of this dimension
    float[] pDimValues = pData.getDimensionValues(m_iDimId);
    
    
    //3. if the specified dimID does not exist
    if (pDimValues == null) {

      //fill the renderColor array with selected color
      m_pRenderColors = new Color[iEntryNum];

      for (int i = 0; i < iEntryNum; i++) {
        m_pRenderColors[i] = m_pMaxColor;
      }

      m_bUpdateColoring = false;

      return;


    }



    //4. calculate the RENDER COLORs
    float fMinValue = pData.getDimensionMinimum(m_iDimId);
    float fMaxValue = pData.getDimensionMaximum(m_iDimId);
    float fNormValue = fMaxValue - fMinValue;

    //components of the selected color if a simple linear transition is used
    float fRed1 = (float)m_pMaxColor.getRed();
    float fGreen1 = (float)m_pMaxColor.getGreen();
    float fBlue1 = (float)m_pMaxColor.getBlue();

    //components of the not selected color
    float fRed2 = (float)m_pMinColor.getRed();
    float fGreen2 = (float)m_pMinColor.getGreen();
    float fBlue2 = (float)m_pMinColor.getBlue();

    m_pRenderColors = new Color[iEntryNum];

    for (int i = 0; i < iEntryNum; i++) {
      float fCurrval = pDimValues[i];

      if (fCurrval <= fMinValue) {
        m_pRenderColors[i] = m_pMinColor;
      }
      else if (fCurrval >= fMaxValue) {
        m_pRenderColors[i] = m_pMaxColor;
      }
      else {
        float fFactor = (fCurrval - fMinValue)/fNormValue;

        if (m_pColorTransition != null) {
          m_pRenderColors[i] = m_pColorTransition.getColorForPos(fFactor);
        }
        else {
          float fInvFactor = 1.0f - fFactor;
          int iRed = (int)(fRed1*fFactor + fInvFactor*fRed2);
          int iGreen = (int)(fGreen1*fFactor + fInvFactor*fGreen2);
          int iBlue = (int)(fBlue1*fFactor + fInvFactor*fBlue2);

          m_pRenderColors[i] = new Color(iRed, iGreen, iBlue);

        }

      }

    }

    //5. set the update flag
    m_bUpdateColoring = false;

  }


}