package view;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import listeners.GlobalConstants;
import listeners.IAxisManagementListener;
import listeners.IExportable;
import listeners.IColoringListener;
import listeners.IViewColoring;

import data.ScatterplotSelection;
import data.Data;
import data.SelectionList;

import tools.ChiSquaredDistribution;

import javax.swing.*;
import java.awt.*;

import java.awt.event.*;
import java.awt.image.*;
import java.util.*;

public class OutlierDetectionView extends JPanel 
  implements MouseListener, MouseMotionListener, GlobalConstants, IExportable, IColoringListener, Runnable {



  /*** data specific variables ***/
  private float[] m_pRobustDistances;      //robust distance array
  private float[] m_pMahalanobisDistances; //mahalanobis distance array
  private float   m_fMaxRobDist;           //maximal robust distance
  private float   m_fMaxMahDist;           //maximal mahalanobis distance

  private Data    m_pData;                //data object needed for the coloring
  private float   m_fQuantile;           //which quantile of the chi squared distribution do we need
  private float   m_fChiSqQuantile;      //quantile value of the chi squared distribution = decision line for outliers
  private int     m_iDF;                 //degree of freedom for the chi squared quantiles


  /*** Images for displaying the view ***/
  private BufferedImage m_pEmptyAxisImage;  //contains the axis and the white background
  private BufferedImage m_pDataImage;       //contains the axis, background and data points

  /*** Flags indicating what is to redraw ***/
  private boolean m_bRedrawAll;             //everything has to be redrawn
  private boolean m_bRedrawData;           //the data has to be redrawn
  private boolean m_bRedrawSelections;     //only the selection borders are redrawn



  /*** drawing variables ***/
  private int     m_iDrawingWidth;  //width in pixels for drawing data points
  private int     m_iDrawingHeight; //height in pixels for drawing data points
  private int     m_iWidth;         //width of the view
  private int     m_iHeight;        //height of the view

  /*** COLORING ***/
  IViewColoring   m_pColoringView; //object managing coloring of data points

  /*** interaction ***/
  Interaction m_pInteraction;     //object handling the interaction functionality for this view



  /*** Thread for redrawing the data items ***/
  Thread m_pRedrawDataThread;      //thread managing the redraw of the data points
  boolean m_bRedrawDataDone;  //indicates if the redraw of the data has been done
  private java.util.Timer m_pTimer;   //initiates a new drawing of the data points




  ///////////////////////////
  // CONSTRUCTOR           //
  ///////////////////////////
  public OutlierDetectionView(int iWidth, int iHeight) {
    //1. set view parameters
    setSize(iWidth, iHeight);

    //2. add Mouse interface functionality
    addMouseListener(this);
    addMouseMotionListener(this);

    //3. set members 
    m_pEmptyAxisImage = null;
    m_pRobustDistances = null;
    m_pMahalanobisDistances = null;
    

    //3. set coloring, interaction
    m_pColoringView = new SelectionColoring(Color.BLACK, Color.RED, null);
    m_pInteraction = new OutlierDetectionInteraction(new Vector(), this);


    //4. set thread variables
    m_bRedrawDataDone = true;


    //5. set data variables
    m_pData = null;
    m_iDF = 2;
    setOutlierPercentage(0.975f);



  }



  ///////////////////////////////
  // PAINTING                  //
  ///////////////////////////////
  
  //calls the paint-method
  public void paintComponent(Graphics pGraphics) {
    super.paintComponent(pGraphics);
    update(pGraphics);
  }


  //paint-method
  public void update(Graphics pGraphics) {
    //1. make some checks
    if (m_pRobustDistances == null) return;
    if (m_pMahalanobisDistances == null) return;
  
    //2. configure the graphics context
    Graphics2D pGraphics2D = (Graphics2D)pGraphics;
    pGraphics2D.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    pGraphics2D.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);

    


    //3. check if size has changed
    if ((m_pEmptyAxisImage == null) ||
        (m_pEmptyAxisImage.getWidth() != getWidth()) ||
        (m_pEmptyAxisImage.getHeight() != getHeight())) {

      m_bRedrawAll = true;
      m_pDataImage = null;
    }

    
    //4. if necessary redraw the background image
    if (m_bRedrawAll) {
      redrawBackgroundImage();
    }

    //5. if necessary redraw the data image
    if (m_bRedrawAll || m_bRedrawData) {
      if (m_pDataImage == null) {
        redrawDataImage();
      }
      else {
        //start the thread for redrawing the data points if
        //the previous redraw thread has been done
        if (m_bRedrawDataDone) {
          m_pRedrawDataThread = new Thread(this);
          m_pRedrawDataThread.start();
          m_bRedrawDataDone = false;
        }
        //cancel the redraw if it is still running
        else {
          if (m_pTimer != null) m_pTimer.cancel();
          m_pTimer = new java.util.Timer();
          m_pTimer.schedule(new RedrawTask(this), 200);
        }
      }

    }


    //6. set the redraw flags
    m_bRedrawAll = false;
    m_bRedrawData = false;

    
    //7. redraw the data image
    pGraphics2D.drawImage(m_pDataImage, 0, 0, this);

    //8. redraw the outlier borders
    redrawOutlierBorder(pGraphics2D);


  }




  ////////////////////////////////////////
  //DRAWING METHODS                     //
  ////////////////////////////////////////

  //redraws the image m_pEmptyAxisImage that contains the axis and the white background
  private void redrawBackgroundImage() {
    
    //1. create the image and the graphics context
    int iWidth = getWidth();
    int iHeight = getHeight();

    BufferedImage pImage = new BufferedImage(iWidth, iHeight, BufferedImage.TYPE_INT_RGB);
    Graphics2D pGraphics = pImage.createGraphics();
    pGraphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    pGraphics.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);

    //2. draw the background
    pGraphics.setColor(Color.WHITE);
    pGraphics.fillRect(0, 0, iWidth, iHeight);


    //3. draw the diagonal
    int iX1 = ARROW_SPACE;
    int iY1 = iHeight - ARROW_SPACE;
    int iX2, iY2;

    if (m_fMaxRobDist > m_fMaxMahDist) {
      iX2 = iWidth - ARROW_TOP_LENGTH;

      int iDrawingHeight = iHeight - ARROW_TOP_LENGTH - ARROW_SPACE;
      iY2 = (int)(iHeight - ARROW_SPACE - (m_fMaxMahDist/m_fMaxRobDist)*(float)(iDrawingHeight));
    }
    else {
      iY2 = ARROW_TOP_LENGTH;

      int iDrawingWidth = iWidth - ARROW_TOP_LENGTH - ARROW_SPACE;
      iX2 = (int)(ARROW_SPACE + (m_fMaxRobDist/m_fMaxMahDist)*(float)(iDrawingWidth));
    }

    pGraphics.setStroke(new BasicStroke(2.0f));
    pGraphics.setColor(Color.GREEN);
    pGraphics.drawLine(iX1, iY1, iX2, iY2);




    //4. draw x-axis arrow

    //4.1. x-axis:rectangle
    pGraphics.setColor(new Color(0, 155, 0));

    int iXStart = (ARROW_SPACE - ARROW_THICKNESS)/2;
    int iYStart = iHeight - (ARROW_SPACE - ARROW_THICKNESS)/2 - ARROW_THICKNESS;
    int iArrowWidth = iWidth - iXStart - ARROW_TOP_LENGTH;
    pGraphics.fillRect(iXStart, iYStart, iArrowWidth, ARROW_THICKNESS);

    //4.2. x-axis: dimension name
    pGraphics.setFont(new Font("Arial", Font.BOLD, 15));
    FontMetrics pXFontMetrics = pGraphics.getFontMetrics();
    int iStringLength = pXFontMetrics.stringWidth("Mahalanobis distance");

    int iStringStart = (iArrowWidth - iStringLength)/2 + (ARROW_SPACE - ARROW_THICKNESS)/2;
    pGraphics.drawString("Mahalanobis distance", iStringStart, iHeight-5);

    //4.3. x-axis: minimum
    pGraphics.drawString("0", ARROW_SPACE, iHeight-5);

    //4.4. x-axis: maximum
    float fXMax = (float)((int)(m_fMaxMahDist*100))/100;
    iStringStart = iWidth - ARROW_TOP_LENGTH - pXFontMetrics.stringWidth(fXMax+"");
    pGraphics.drawString(fXMax+"", iStringStart, iHeight-5);



    //4.5. x-axis: arrow top
    Polygon pArrowTop = new Polygon();
    int iXCoord = iXStart + iArrowWidth;
    int iYCoord = iHeight - (ARROW_SPACE - ARROW_TOP_WIDTH)/2;
    pArrowTop.addPoint(iXCoord,iYCoord);
    
    iYCoord = iHeight - (ARROW_SPACE - ARROW_TOP_WIDTH)/2 - ARROW_TOP_WIDTH;
    pArrowTop.addPoint(iXCoord,iYCoord);
    
    iXCoord = iWidth-1;
    iYCoord = iHeight - ARROW_SPACE/2;
    pArrowTop.addPoint(iXCoord,iYCoord);

    pGraphics.fill(pArrowTop);





    //5. draw y-axis arrow
    //5.1. y-axis:rectangle
    pGraphics.setColor(new Color(0, 0, 155));

    iXStart = (ARROW_SPACE - ARROW_THICKNESS)/2;
    iYStart = ARROW_TOP_LENGTH;
    int iArrowHeight = iHeight - (ARROW_SPACE - ARROW_THICKNESS)/2 - ARROW_TOP_LENGTH;
    pGraphics.fillRect(iXStart, iYStart, ARROW_THICKNESS, iArrowHeight);

    //5.2. y-axis: arrow top
    pArrowTop = new Polygon();
    iXCoord = (ARROW_SPACE - ARROW_TOP_WIDTH)/2;
    iYCoord = ARROW_TOP_LENGTH;
    pArrowTop.addPoint(iXCoord,iYCoord);
    
    iXCoord = (ARROW_SPACE - ARROW_TOP_WIDTH)/2 + ARROW_TOP_WIDTH;
    pArrowTop.addPoint(iXCoord,iYCoord);
    
    iXCoord = ARROW_SPACE/2;
    iYCoord = 0;
    pArrowTop.addPoint(iXCoord,iYCoord);

    pGraphics.fill(pArrowTop);

    
    //5.3. y-axis: dimension name
    pGraphics.rotate(Math.PI/2);
    iStringLength = pXFontMetrics.stringWidth("Distances");
    iStringStart = ARROW_TOP_LENGTH + (iArrowHeight-iStringLength)/2;

    pGraphics.drawString("Distances", iStringStart, -5);


    //5.4. y-axis: maximum
    float fYMax = (float)((int)(m_fMaxRobDist*100))/100;
    pGraphics.drawString(fYMax+"", ARROW_TOP_LENGTH, -5);

    //5.5. y-axis: minimum
    iStringStart = iHeight - ARROW_SPACE - pXFontMetrics.stringWidth("0");
    pGraphics.drawString("0", iStringStart, -5);



    //6. set m_pEmptyAxisImage
    m_pEmptyAxisImage = pImage;


  }


  //redraws the image m_pDataImage that is drawn on basis of m_pEmptyAxisImage
  //here the data points are drawn
  private void redrawDataImage() {

    //1. calculate drawing variables
    m_iWidth = getWidth();
    m_iDrawingWidth = m_iWidth - ARROW_SPACE - ARROW_TOP_LENGTH;
    m_iHeight = getHeight();
    m_iDrawingHeight = m_iHeight - ARROW_SPACE - ARROW_TOP_LENGTH;


    //2. create a new data image
    BufferedImage pImage = new BufferedImage(m_iWidth, m_iHeight, BufferedImage.TYPE_INT_RGB);
    Graphics2D pGraphics = pImage.createGraphics();
    pGraphics.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
    pGraphics.setRenderingHint(RenderingHints.KEY_RENDERING, RenderingHints.VALUE_RENDER_SPEED);

    //3. draw the background
    pGraphics.drawImage(m_pEmptyAxisImage, 0, 0, null);

    //4. draw the data points
    Color[] pDataPointColors = m_pColoringView.getRenderColors(m_pData);
    int[] pRenderPositions = m_pColoringView.getRenderPositions(m_pData);
    int iEntryNum = m_pRobustDistances.length;
    
    if (pRenderPositions == null) {

      for (int i = 0; i < iEntryNum; i++) {
        pGraphics.setColor(pDataPointColors[i]);

        float fXPos = (float)m_pMahalanobisDistances[i]/m_fMaxMahDist;
        fXPos = ARROW_SPACE + (fXPos * m_iDrawingWidth);

        float fYPos = m_pRobustDistances[i]/m_fMaxRobDist;
        fYPos = m_iHeight - ARROW_SPACE - (fYPos * m_iDrawingHeight);

        pGraphics.fillOval((int)fXPos-2, (int)fYPos-2, 4, 4);

      }
    }
    else {
      
      for (int i = 0; i < iEntryNum; i++) {
        pGraphics.setColor(pDataPointColors[i]);
        int id = pRenderPositions[i];

        float fXPos = (float)m_pMahalanobisDistances[i]/m_fMaxMahDist;
        fXPos = ARROW_SPACE + (fXPos * m_iDrawingWidth);

        float fYPos = m_pRobustDistances[i]/m_fMaxRobDist;
        fYPos = m_iHeight - ARROW_SPACE - (fYPos * m_iDrawingHeight);

        pGraphics.fillOval((int)fXPos-2, (int)fYPos-2, 4, 4);

      }
    }

    //set the image
    m_pDataImage = pImage;
    m_bRedrawDataDone = true;

  }


  //draws outlier borders
  private void redrawOutlierBorder(Graphics2D pGraphics) {
    //1. make a simple check
    if (pGraphics == null) return;
    
    //2. make the basic graphics settings
    pGraphics.setColor(new Color(0, 150, 150));
    pGraphics.setStroke(new BasicStroke(4.0f));

    //3. draw the horizontal border
    int iX1 = ARROW_SPACE;
    int iX2 = ARROW_SPACE + m_iDrawingWidth;
    int iY = getYPixelPosition(m_fChiSqQuantile);

    pGraphics.drawLine(iX1, iY, iX2, iY);


    //4. draw the quantile description
    float fQuantVal = (float)((int)(m_fChiSqQuantile*1000))/1000;
    pGraphics.setFont(new Font("Arial", Font.BOLD, 15));
    pGraphics.drawString(fQuantVal+"", iX1, iY - 2);
    pGraphics.drawString("DF = " + m_iDF, iX1, iY - 17);


    //5. vertical border
    int iY1 = m_iHeight - ARROW_SPACE;
    int iY2 = ARROW_TOP_LENGTH;
    int iX = getXPixelPosition(m_fChiSqQuantile);

    pGraphics.drawLine(iX, iY1, iX, iY2);



  }




  ////////////////////////////////////////
  //Coordinate calculations             //
  ////////////////////////////////////////

  // calculates the x pixel position for a certain entry
  public int getXPixelPosition(float fValue) {
    float fXPos = (float)fValue/m_fMaxMahDist;
    fXPos = ARROW_SPACE + (fXPos * m_iDrawingWidth);

    return (int)fXPos;
  }

  // calculates the pixel position for a certain value on the y axes
  public int getYPixelPosition(float fValue) {
    float fYPos = (float)fValue/m_fMaxRobDist;
    fYPos = m_iHeight - ARROW_SPACE - (fYPos * m_iDrawingHeight);

    return (int)fYPos;

  }







  ////////////////////////////////////////
  //MouseListener methods               //
  ////////////////////////////////////////
  //mouse clicked: show selection menu
  public void mouseClicked(MouseEvent e) {
    m_pInteraction.mouseClicked(e);
  }

  public void mouseEntered(MouseEvent e) {
    m_pInteraction.mouseEntered(e);
  }

  public void mouseExited(MouseEvent e) {
    m_pInteraction.mouseExited(e);
  }

  public void mousePressed(MouseEvent e) {
    m_pInteraction.mousePressed(e);
  }

  public void mouseReleased(MouseEvent e) {
    m_pInteraction.mouseReleased(e);
  }


  ////////////////////////////////////////
  //MouseMotionListener methods         //
  ////////////////////////////////////////
  public void mouseDragged(MouseEvent e) {
    m_pInteraction.mouseDragged(e);
  }

  public void mouseMoved(MouseEvent e) {
    m_pInteraction.mouseMoved(e);
  }





  //////////////////////////////////////////
  // interface for the key events         //
  // triggered from the mainframe         //
  //////////////////////////////////////////

  public void setControlPressed(boolean pressed) {
    m_pInteraction.setControlPressed(pressed);
  }



  //////////////////////////////////////////
  // GET METHODS                          //
  //////////////////////////////////////////



  public void renderImage(BufferedImage pImage) {
    //1. make a simple check
    if (pImage == null) return;
    
    //2. save old size
    int iOldWidth = getWidth();
    int iOldHeight = getHeight();

    //3. paint on the image and set the size
    Graphics2D pGraphics = pImage.createGraphics();
    int iWidth = pImage.getWidth();
    int iHeight = pImage.getHeight();
    setSize(iWidth, iHeight);
    
    m_bRedrawAll = true;
    paint(pGraphics);

    //4. reset the size
    setSize(iOldWidth, iOldHeight);


  }


  public float[] getDOIofOutliers() {
    int iEntryNum = m_pRobustDistances.length;
    float[] pDOI = new float[iEntryNum];

    for (int i = 0; i < iEntryNum; i++) {
      if (m_pRobustDistances[i] > m_fChiSqQuantile) {
        pDOI[i] = 1.0f;
      }
      else {
        pDOI[i] = 0.0f;
      }
    }

    return pDOI;
  }

  
  public float[] getDOIofNonOutliers() {
    int iEntryNum = m_pRobustDistances.length;
    float[] pDOI = new float[iEntryNum];

    for (int i = 0; i < iEntryNum; i++) {
      if (m_pRobustDistances[i] > m_fChiSqQuantile) {
        pDOI[i] = 0.0f;
      }
      else {
        pDOI[i] = 1.0f;
      }
    }

    return pDOI;
  }



  //////////////////////////////////////////
  // SET METHODS                          //
  //////////////////////////////////////////





  //sets a new dataset
  public void setDistances(float[] pRobustDistances, float[] pMahalanobisDistances, int iDegreeOfFreedom) {
    //1. make some checks
    if (pRobustDistances == null) return;
    if (pMahalanobisDistances == null) return;
    if (pRobustDistances.length != pMahalanobisDistances.length) return;

    //2. set the members
    m_pRobustDistances = pRobustDistances;
    m_pMahalanobisDistances = pMahalanobisDistances;
    m_iDF = iDegreeOfFreedom;
    
    
    //3. calculate the maximum distances
    int iEntryNum = m_pRobustDistances.length;
    m_fMaxRobDist = 0;
    m_fMaxMahDist = 0;
    for (int i = 0; i < iEntryNum; i++) {
      if (m_pRobustDistances[i] > m_fMaxRobDist) {
        m_fMaxRobDist = m_pRobustDistances[i];
      }

      if (m_pMahalanobisDistances[i] > m_fMaxMahDist) {
        m_fMaxMahDist = m_pMahalanobisDistances[i];
      }

    }

    
    //4. set the outlier percentage
    setOutlierPercentage(m_fQuantile);

    //5. do the painting
    m_bRedrawAll = true;
    repaint();
  }


  public void setData(Data pData) {
    m_pData = pData;
  }


  public void setInteraction(Interaction pInteraction) {
    m_pInteraction = pInteraction;
  }

  public void setOutlierPercentage(float fPercentage) {

    m_fQuantile = fPercentage;
    m_fChiSqQuantile = (float)Math.sqrt(ChiSquaredDistribution.getQuantileValue(m_iDF, (double)fPercentage));

    repaintBorder();
  }



  //////////////////////////////////////////
  // CALLS FOR REPAINTING                 //
  //////////////////////////////////////////
  public void repaintBorder() {
    repaint();
  }

  public void repaintData() {
    m_bRedrawData = true;
    repaint();
  }


  //initializes a redraw of the view. redrawBackground indicates if all data items should be redrawn too.
  public void repaintAll() {
    m_bRedrawAll = true;
    repaint();
  }


  //////////////////////////////////////////
  // ICOLORINGLISTENER METHODS            //
  //////////////////////////////////////////

  /** Invoked when a change in the coloring of the data items occurs.
   *
   *  @param coloring IViewColoring object managing the colors of the data items.
   */
  public void coloringChanged(IViewColoring pColoringView) {
    //1. make a simple check
    if (pColoringView == null) return;
    
    //2. set the member
    m_pColoringView = pColoringView;

    //3. schedule a redraw
    if (m_pTimer != null) m_pTimer.cancel();
    m_pTimer = new java.util.Timer();
    m_pTimer.schedule(new RedrawTask(this), 200);

    repaintData();
  }


  //////////////////////////////////////////
  // RUNNABLE INTERFACE                   //
  //////////////////////////////////////////
  public void run() {
    redrawDataImage();
  }


}