package view;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */

import data.AxisSelection;
import data.SelectionList;
import data.Selection;
import listeners.GlobalConstants;
import listeners.IDataSubsetListener;

import javax.swing.*;
import java.awt.event.*;
import java.awt.Cursor;
import java.util.Vector;


/* class handling the interactive creation and manipulation of selections for the parallel coordinate view */
public class ParCoordSelectionInteraction extends Interaction implements GlobalConstants, ActionListener {


  /** member variables **/
  private JPopupMenu    m_pAxisSelectionPopupMenu;    // popup menu for the AxisSelections
  private ParCoordView  m_pParCoordView; // the parallel coordinates view

  /** MOUSE STATUS **/
  int m_iMouseStatus;       //is mouse in, out or on an horizontal border of a selection
  int m_iMouseDragStatus;   //should the mouse resize or reposition a selection or should a soft selection be created


  /** helpers for the selction manipulation **/
  float   m_fStartBorder;           //for the creation of a axisselection:
                                    //indicating what the starting border of the selection is
  int     m_iRepositionY;           //needed for selection repositioning and resizing: holds the last y position
  boolean m_bResizeUpperBorder;     //needed for resizing: should upperborder be changed?
  boolean m_bResizeSoftBorder;      //should the soft border be changed?
  boolean m_bCreateSoftBorder;      //should be a soft selection created?
  boolean m_bFlippedAxis;           //indicates of the current axis is flipped.

  /**** key variables ****/
  boolean m_bCtrlPressed;       //indicates if control key is pressed (needed for drawing soft selection)

  /** selections **/
  AxisSelection m_pCurrentAxisSelection; //axis selection that is currently modified by the mouse
  SelectionList m_pSelectionList;            //object managing the DOI and all selections



  //////////////////////////////////
  // CONSTRUCTOR                  //
  //////////////////////////////////
  /** constructor **/
  public ParCoordSelectionInteraction(
    Vector<IDataSubsetListener> pListeners, 
    ParCoordView pParCoordView, 
    SelectionList pSelectionList
  ) 
  {
    //1. call the father constructor
    super(pListeners);

    //2. set member variables
    m_pParCoordView = pParCoordView;
    m_pSelectionList = pSelectionList;

    //3. create the popup menus
    createAxisSelectionPopupMenu();


    //4. set default mouse status
    m_iMouseStatus = MOUSE_OUT;
    m_iMouseDragStatus = MOUSE_DRAG_NONE;

    //5. set default helpers status
    m_bCreateSoftBorder = false;

    //6. set default key status
    m_bCtrlPressed = false;

    //7. set default selection status
    m_pCurrentAxisSelection = null;
    m_bFlippedAxis = false;



  }



  ////////////////////////////////////////
  //MOUSE LISTENER METHODS              //
  ////////////////////////////////////////
  //mouse clicked: show selection menu
  public void mouseClicked(MouseEvent e) {
    //1. don't do anything, when it is no right mouse click
    if (e.getButton() != MouseEvent.BUTTON3) return;

    //2. show the popup menu, when the mouse is within a selection or at the border of a selection
    if ((m_iMouseStatus == MOUSE_IN) || (m_iMouseStatus == MOUSE_HORIZONTAL))  {
      m_pAxisSelectionPopupMenu.show(m_pParCoordView, e.getX(), e.getY());
    }

  }


  
  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e) {}


  //mouse button pressed
  public void mousePressed(MouseEvent e) {

    //1. do nothing if right mouse click was used
    if (e.getButton() == MouseEvent.BUTTON3) return;

    //2. reposition the current selection when the mouse is within a selection and the current selection is set
    if ((m_iMouseStatus == MOUSE_IN) && (m_pCurrentAxisSelection != null)) {
      //set the status variables
      m_iMouseDragStatus = MOUSE_DRAG_REPOSITION;
      m_iRepositionY = e.getY();
      return;
    }


    //3. change the border of the current selection or create a soft border
    if ((m_iMouseStatus == MOUSE_HORIZONTAL) && (m_pCurrentAxisSelection != null)) {

      //3.1. set the status variables
      m_iMouseDragStatus = MOUSE_DRAG_RESIZE;
      m_iRepositionY = e.getY();

      //3.2. calculate which border should be resized
      int iDimId = m_pCurrentAxisSelection.getDimId();
      int iUpperdiff = m_iRepositionY - m_pParCoordView.getPixelPosition(iDimId, m_pCurrentAxisSelection.getUpperBorder());
      int iLowerdiff = m_iRepositionY - m_pParCoordView.getPixelPosition(iDimId, m_pCurrentAxisSelection.getLowerBorder());

      //upper or lower border
      m_bResizeUpperBorder = (Math.abs(iUpperdiff) < Math.abs(iLowerdiff));
      m_bResizeSoftBorder = false;
      int iMinDiff = Math.min(Math.abs(iUpperdiff), Math.abs(iLowerdiff));

      //upper soft border
      if (m_pCurrentAxisSelection.hasSoftUpperBorder()) {
        int iSoftUpperDiff = m_iRepositionY - 
        m_pParCoordView.getPixelPosition(iDimId, m_pCurrentAxisSelection.getSoftUpperBorder());

        if (Math.abs(iSoftUpperDiff) < iMinDiff) {
          iMinDiff = Math.abs(iSoftUpperDiff);
          m_bResizeUpperBorder = true;
          m_bResizeSoftBorder = true;
        }
      }
      //lower soft border
      if (m_pCurrentAxisSelection.hasSoftLowerBorder()) {
        int iSoftLowerDiff = m_iRepositionY - 
        m_pParCoordView.getPixelPosition(iDimId, m_pCurrentAxisSelection.getSoftLowerBorder());

        if (Math.abs(iSoftLowerDiff) < iMinDiff) {
          iMinDiff = Math.abs(iSoftLowerDiff);
          m_bResizeUpperBorder = false;
          m_bResizeSoftBorder = true;
        }
      }

      //3.3. if control is pressed and there is no soft border yet -> a soft border is created
      if (m_bResizeUpperBorder) {
        m_bCreateSoftBorder = m_bCtrlPressed && (!m_pCurrentAxisSelection.hasSoftUpperBorder());
      }
      else {
        m_bCreateSoftBorder = m_bCtrlPressed && (!m_pCurrentAxisSelection.hasSoftLowerBorder());
      }

      return;
    }

    //4. create a new selection
    if ((m_iMouseStatus == MOUSE_OUT) && (m_pCurrentAxisSelection == null)) {

      int iXPos = e.getX();
      int iYPos = e.getY();

      int iDimId = m_pParCoordView.calculateDimID(iXPos);
      if (iDimId < 0) return;

      m_iMouseDragStatus = MOUSE_DRAG_CREATE;
      m_fStartBorder = m_pParCoordView.getCorrespondingValue(iDimId, iYPos);
      m_pCurrentAxisSelection = new AxisSelection(iDimId, m_fStartBorder, m_fStartBorder, VIEW_ID_PAARCOORD);
      m_pSelectionList.addSelection(m_pCurrentAxisSelection);

      return;

    }
  }



  //mouse is released: update all selection settings and initialize a recalculation of the view
  public void mouseReleased(MouseEvent e) {
    if (m_iMouseDragStatus != MOUSE_DRAG_NONE) {
      mouseDragged(e);
      //recalculateDOI(); //TO DO: handle this one
      m_iMouseDragStatus = MOUSE_DRAG_NONE;
      m_pParCoordView.repaintSelections();
    }
  }


  ////////////////////////////////////////
  //MOUSE MOTION LISTENER METHODS       //
  ////////////////////////////////////////
  public void mouseDragged(MouseEvent e) {
  
    //1. make some checks
    if (m_pCurrentAxisSelection == null) return;

    //2. repositioning of a selection
    if (m_iMouseDragStatus == MOUSE_DRAG_REPOSITION) {

      //2.1. calculate the difference on the screen
      int iYPos = e.getY();
      int iYDifference = iYPos - m_iRepositionY;
      if (m_bFlippedAxis) {
        iYDifference = -iYDifference;
      }

      //2.2. calculate the difference on the dimension and set it in the selection
      int iDimId = m_pCurrentAxisSelection.getDimId();
      float fValueDifference = m_pParCoordView.calculateCorrespondingDistanceValue(iDimId, iYDifference);

      m_pCurrentAxisSelection.setUpperBorder(fValueDifference + m_pCurrentAxisSelection.getUpperBorder());
      m_pCurrentAxisSelection.setLowerBorder(fValueDifference + m_pCurrentAxisSelection.getLowerBorder());

      if (m_pCurrentAxisSelection.hasSoftUpperBorder()) {
        m_pCurrentAxisSelection.setSoftUpperBorder(fValueDifference + m_pCurrentAxisSelection.getSoftUpperBorder());
      }
      if (m_pCurrentAxisSelection.hasSoftLowerBorder()) {
        m_pCurrentAxisSelection.setSoftLowerBorder(fValueDifference + m_pCurrentAxisSelection.getSoftLowerBorder());
      }

      //2.3. repaint the selections and set the Y Position
      m_pParCoordView.repaintSelections();
      m_iRepositionY = iYPos;

      return;

    }

    //3. change the border of the current selection or create a soft border
    if (m_iMouseStatus == MOUSE_HORIZONTAL) {

      //3.1. calculate the difference on the screen
      int iYPos = e.getY();
      int iYDifference = iYPos - m_iRepositionY;
      if (m_bFlippedAxis) {
        iYDifference = -iYDifference;
      }

      //3.2. calculate the value difference on the dimension
      int iDimId = m_pCurrentAxisSelection.getDimId();
      float fValueDifference = m_pParCoordView.calculateCorrespondingDistanceValue(iDimId, iYDifference);

      //3.3. create or resize the soft border
      boolean bIsValueSet;
      if (m_bCreateSoftBorder || m_bResizeSoftBorder) {

        if (m_bResizeUpperBorder) {
          bIsValueSet = m_pCurrentAxisSelection.setSoftUpperBorder(fValueDifference + m_pCurrentAxisSelection.getSoftUpperBorder());
        }
        else {
          bIsValueSet = m_pCurrentAxisSelection.setSoftLowerBorder(fValueDifference + m_pCurrentAxisSelection.getSoftLowerBorder());
        }
      }
      //3.4. change the hard border
      else {

        if (m_bResizeUpperBorder) {
          bIsValueSet = m_pCurrentAxisSelection.setUpperBorder(fValueDifference + m_pCurrentAxisSelection.getUpperBorder());
        }
        else {
          bIsValueSet = m_pCurrentAxisSelection.setLowerBorder(fValueDifference + m_pCurrentAxisSelection.getLowerBorder());
        }
      }

      
      //3.5. repaint the selections and make some settings
      m_pParCoordView.repaintSelections();

      //only if the value has been set a new m_iRepositionY is stored!
      if (bIsValueSet) m_iRepositionY = iYPos;

      return;

    }


    //4. creating a new selection
    if (m_iMouseDragStatus == MOUSE_DRAG_CREATE) {
      int iYPos = e.getY();
      int iDimId = m_pCurrentAxisSelection.getDimId();

      float fBorder = m_pParCoordView.getCorrespondingValue(iDimId, iYPos);

      m_pSelectionList.deleteSelection(m_pCurrentAxisSelection, false);

      m_pCurrentAxisSelection = new AxisSelection(iDimId, m_fStartBorder, fBorder, VIEW_ID_PAARCOORD);
      m_pSelectionList.addSelection(m_pCurrentAxisSelection);

      m_pParCoordView.repaintSelections();
      return;

    }

  }

  public void mouseMoved(MouseEvent e) {
    //1. make a simple check
    if (m_iMouseDragStatus != MOUSE_DRAG_NONE) return;

    //2. recalculate the mouse status
    recalculateMouseStatus(e.getX(), e.getY());

    //3. set the mouse cursor according to the mouse status
    switch (m_iMouseStatus) {
      case MOUSE_OUT: {
        m_pParCoordView.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        break;
      }
      case MOUSE_HORIZONTAL: {
        m_pParCoordView.setCursor(new Cursor(Cursor.N_RESIZE_CURSOR ));
        break;
      }
      case MOUSE_IN: {
        m_pParCoordView.setCursor(new Cursor(Cursor.MOVE_CURSOR));
        break;
      }

    }


  }


  //////////////////////////////////////////
  // interface for the key events         //
  // triggered from the mainframe         //
  //////////////////////////////////////////
  public void setControlPressed(boolean bPressed) {
    m_bCtrlPressed = bPressed;
  }




  // calculates the status of the a mouse position and sets the current AxisSelection
  // 0 = outside of AxisSelections
  // 1 = on horizontal border of a selection
  // 2 = in the selection
  public void recalculateMouseStatus(int iXPos, int iYPos) {

    //1. get the dimension id according to the x position
    int iDimId = m_pParCoordView.calculateDimID(iXPos);
    //System.out.println("dimID: " + dimID);
    if (iDimId < 0) return;

    //2. check the x position (if the mouse is out of the axis borders, set 
    //   mouse status = OUT and no current axis selection
    int[] pBorders = m_pParCoordView.getHorizontalSelectionBorders(iDimId);
    if (pBorders == null) return;
    int iXBorderLow = pBorders[0];
    int iXBorderUp = pBorders[1];

    if (iXPos < iXBorderLow) {
      m_iMouseStatus = MOUSE_OUT;
      m_pCurrentAxisSelection = null;
      return;
    }
    if (iXPos > iXBorderUp) {
      m_iMouseStatus = MOUSE_OUT;
      m_pCurrentAxisSelection = null;
      return;
    }


    //3. ask if the axis is flipped
    m_bFlippedAxis = m_pParCoordView.isAxisFlipped(iDimId);

    //4. fetch selections
    Vector<Selection> pSelections = m_pSelectionList.getLocalSelections(VIEW_ID_PAARCOORD);
    int iSelectionNum = pSelections.size();


    //5. check the y position
    for (int i = 0; i < iSelectionNum; i++) {
      //the current selection is not at the correct axis -> skip it
      AxisSelection pCurrSelection = (AxisSelection)pSelections.get(i);
      if (pCurrSelection.getDimId() != iDimId) continue;

      //get the soft borders: they are at least at the positions of the hard borders
      int iUpperBorder = m_pParCoordView.getPixelPosition(iDimId, pCurrSelection.getSoftUpperBorder());
      int iLowerBorder = m_pParCoordView.getPixelPosition(iDimId, pCurrSelection.getSoftLowerBorder());

      //flipped axis: change lower and upperborder
      if (m_bFlippedAxis) {
        int help = iLowerBorder;
        iLowerBorder = iUpperBorder;
        iUpperBorder = help;
      }


      //check if the mouse is within the upper and lower border plus a specific margin
      if ((iYPos > iUpperBorder-5) && (iYPos < iLowerBorder+5)) {

        //mouse is within the soft borders
        if ((iYPos > iUpperBorder+5) && (iYPos < iLowerBorder-5)) {

          //check the hard borders
          int iUpperHardBorder = m_pParCoordView.getPixelPosition(iDimId, pCurrSelection.getUpperBorder());
          int iLowerHardBorder = m_pParCoordView.getPixelPosition(iDimId, pCurrSelection.getLowerBorder());

          //mouse is on hard borders
          if (((iYPos > iUpperHardBorder-5) && (iYPos < iUpperHardBorder+5)) ||
              ((iYPos > iLowerHardBorder-5) && (iYPos < iLowerHardBorder+5))) {
            m_iMouseStatus = MOUSE_HORIZONTAL;
            m_pCurrentAxisSelection = pCurrSelection;
            return;
          }
          
          // the mouse is within a selection
          m_iMouseStatus = MOUSE_IN;
          m_pCurrentAxisSelection = pCurrSelection;
          return;
        }
        //mouse is on a horizontal border
        else {
          m_iMouseStatus = MOUSE_HORIZONTAL;
          m_pCurrentAxisSelection = pCurrSelection;
          return;
        }
      }




    }

    //6. set the default status if the mouse was not near to a selection
    m_iMouseStatus = MOUSE_OUT;
    m_pCurrentAxisSelection = null;

  }




  ////////////////////////////////////////
  //POPUP MENUS                         //
  ////////////////////////////////////////
  private void createAxisSelectionPopupMenu() {

    m_pAxisSelectionPopupMenu = new JPopupMenu();

    JMenuItem pDeleteSelection = new JMenuItem("Delete selection");
    pDeleteSelection.addActionListener(this);
    m_pAxisSelectionPopupMenu.add(pDeleteSelection);

    JMenuItem pDeleteAllSelections = new JMenuItem("Delete all selections");
    pDeleteAllSelections.addActionListener(this);
    m_pAxisSelectionPopupMenu.add(pDeleteAllSelections);

    JMenuItem pSetSelectedDataAsData = new JMenuItem("Work with selected data");
    pSetSelectedDataAsData.addActionListener(this);
    m_pAxisSelectionPopupMenu.add(pSetSelectedDataAsData);

    JMenuItem pResetWholeDatasetAsData = new JMenuItem("Work with whole dataset");
    pResetWholeDatasetAsData.addActionListener(this);
    m_pAxisSelectionPopupMenu.add(pResetWholeDatasetAsData);

  }


  //////////////////////////////////////////
  //ActionListeners: for popup menu items //
  //////////////////////////////////////////
  public void actionPerformed(ActionEvent e) {

    if (e.getActionCommand().equals("Delete selection")) {
      if (m_pCurrentAxisSelection != null) {
        m_pSelectionList.deleteSelection(m_pCurrentAxisSelection, true);
        m_pCurrentAxisSelection = null;
        m_pParCoordView.repaintSelections();
      }
      return;
    }

    if (e.getActionCommand().equals("Delete all selections")) {
      m_pSelectionList.deleteAllLocalSelections(VIEW_ID_PAARCOORD);
      m_pCurrentAxisSelection = null;
      m_pParCoordView.repaintSelections();
      return;
    }

    if (e.getActionCommand().equals("Work with selected data")) {
      setSelectedDataAsData();
      return;
    }

    if (e.getActionCommand().equals("Work with whole dataset")) {
      resetWholeDatasetAsData();
      return;
    }


  }



}