
package view;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import listeners.IViewColoring;
import data.Data;

import java.awt.Color;


public class SelectionColoring implements IViewColoring {


  /*** COLORS ***/
  private   Color       m_pSelectedColor;        //Color for the selected items
  private   Color       m_pNotSelectedColor;     //Color for the not selected items


  /*** COLORTRANSITION ***/
  private   ColorTransition m_pSoftSelectionColorTransition;    //ColorTransition for the soft selection


  /*** RENDER POSITIONS ***/
  private   int[]       m_pRenderPositions; //array with data IDs indicating the render positions


  /*** RENDER COLOR  ***/
  private   Color[]     m_pRenderColors;    //array with colors with which the items should be rendered
  private   int         m_iNotSelectedNum; //number of items that are not selected

  /*** FLAGS ***/
  private   boolean     m_bUpdateSelectionColoring;  //indicating if the coloring must be updated
  private   boolean     m_bUpdateSoftSelectionColoring; //indicating that the coloring for the soft selected items must be updated

  /*** DOI ***/
  private float[] m_pDOI;  //float array holding the doi to draw



  /////////////////////////////
  // CONSTRUCTOR             //
  /////////////////////////////
  public SelectionColoring(Color pNotSelectedColor, Color pSelectedColor, ColorTransition pColorTransition) {
    //1. make some checks
    if (pNotSelectedColor == null) pNotSelectedColor = Color.BLACK;
    if (pSelectedColor == null) pSelectedColor = Color.RED;
    
    //2. set the members
    m_pSelectedColor = pSelectedColor;
    m_pNotSelectedColor = pNotSelectedColor;
    m_pSoftSelectionColorTransition = pColorTransition;

    m_bUpdateSelectionColoring = true;
    m_pDOI = null;

    m_pRenderPositions = null;
    m_pRenderColors = null;
    m_iNotSelectedNum = 0;

  }

  
  ///////////////////////////////
  // GET METHODS               //
  ///////////////////////////////

  /* returns an integer array with the IDs of the data items.
     the first ID in the array should be drawn first and appears in the background.
     the last ID in the array should be drawn as last item and appears in the front.*/
  public int[] getRenderPositions(Data pData) {
    //1. make a simple check
    if (pData == null) return null;
  
    //2. do some recalculations
    if ((m_bUpdateSelectionColoring) || (m_pRenderPositions == null)) {
      recalculateSelectionColoring(pData);
    }
    return m_pRenderPositions;

  }


  /* returns an array of colors for the data items ordered by the render position */
  public Color[] getRenderColors(Data pData) {
    //1. make a check
    if (pData == null) return null;
  
    //2. do some recalculation
    if ((m_bUpdateSelectionColoring) || (m_pRenderColors == null)) {
      recalculateSelectionColoring(pData);
    }
    if (m_bUpdateSoftSelectionColoring) {
      recalculateSoftColoring();
    }
    return m_pRenderColors;
  }


  public void updateDOI(float[] pDOI) {
    //1. make a check
    if (pDOI == null) return;
    
    //2. make some settings
    m_bUpdateSelectionColoring = true;
    m_pDOI = pDOI;
  }

  
  public void setColorTransition(ColorTransition pColorTransition) {
    //1. make some checks
    if ((m_pSoftSelectionColorTransition == null) && (pColorTransition == null)) return;
  
    //2. make some settings
    m_pSoftSelectionColorTransition = pColorTransition;
    m_bUpdateSoftSelectionColoring = true;

  }


  /* set the necessary color properties */
  public void setColoringProperties(Color pNotSelected, Color pSelected, ColorTransition pColorTransition) {
    //TO DO: shift functionality from the constructor to this method?
  }


  /////////////////////////////////
  // PRIVATE METHODS             //
  /////////////////////////////////
  //think this concept over. there are some out of index errors too and it is ineficient
  private void recalculateSelectionColoring(Data pData) {
    //1. make some simple checks
    if (pData == null) return;
    int iEntryNum = pData.getEntryNum();
    
    //2. if there is no doi just set the default render positions and the selected color
    //   TO DO: Is it really necessary to allocate those arrays when there is no doi?
    if (m_pDOI == null) {

      //fill the renderPosition array with an ascending render order
      m_pRenderPositions = new int[iEntryNum];

      for (int i = 0; i < iEntryNum; i++) {
        m_pRenderPositions[i] = i;
      }

      //fill the renderColor array with selected color
      m_pRenderColors = new Color[iEntryNum];
      for (int i = 0; i < iEntryNum; i++) {
        m_pRenderColors[i] = m_pSelectedColor;
      }

      m_bUpdateSelectionColoring = false;

      return;


    }


    //3. calculate the RENDER ORDER
    //create ordering according to the doi
    m_pRenderPositions = new int[iEntryNum];
    int iCounter = 0;

    //3.1. paint the not selected items first
    for (int i = 0; i < iEntryNum; i++) {
      if (m_pDOI[i] > 0) continue;

      m_pRenderPositions[iCounter] = i;
      iCounter++;
    }


    //3.2. paint the soft selected items next
    for (int i = 0; i < iEntryNum; i++) {
      if ((m_pDOI[i] > 0) && (m_pDOI[i] < 1)) {
        m_pRenderPositions[iCounter] = i;
        iCounter++;
      }
    }


    //3.3. paint the full selected items at last
    for (int i = 0; i < iEntryNum; i++) {
      if (m_pDOI[i] < 1) continue;

      m_pRenderPositions[iCounter] = i;
      iCounter++;
    }


    //4. calculate the RENDER COLOR
    m_pRenderColors = new Color[iEntryNum];
    iCounter = 0;

    //4.1. set the colors of the not selected items first
    for (int i = 0; i < iEntryNum; i++) {
      if (m_pDOI[i] > 0) continue;

      m_pRenderColors[iCounter] = m_pNotSelectedColor;
      iCounter++;
    }

    m_iNotSelectedNum = iCounter;



    //4.2. set the colors of the soft selected items
    if (m_pSoftSelectionColorTransition != null) {
      //an advanced colortransition is there
      for (int i = 0; i < iEntryNum; i++) {
        if ((m_pDOI[i] > 0) && (m_pDOI[i] < 1)) {
          m_pRenderColors[iCounter] = m_pSoftSelectionColorTransition.getColorForPos(m_pDOI[i]);
          iCounter++;
        }
      }
    }
    else {
      //there is no advanced color transition

      //components of the selected color if a simple linear transition is used
      float fRed1 = (float)m_pSelectedColor.getRed();
      float fGreen1 = (float)m_pSelectedColor.getGreen();
      float fBlue1 = (float)m_pSelectedColor.getBlue();

      //components of the not selected color
      float fRed2 = (float)m_pNotSelectedColor.getRed();
      float fGreen2 = (float)m_pNotSelectedColor.getGreen();
      float fBlue2 = (float)m_pNotSelectedColor.getBlue();

      for (int i = 0; i < iEntryNum; i++) {
        float fDOI = m_pDOI[i];
        if ((fDOI > 0) && (fDOI < 1)) {
          float fInvDOI = 1 - fDOI;
          int iRed = (int)(fRed1*fDOI + fInvDOI*fRed2);
          int iGreen = (int)(fGreen1*fDOI + fInvDOI*fGreen2);
          int iBlue = (int)(fBlue1*fDOI + fInvDOI*fBlue2);

          m_pRenderColors[iCounter] = new Color(iRed, iGreen, iBlue);
          iCounter++;
        }
      }
    }


    //4.3. set the colors of the selected items at last
    for (int i = 0; i < iEntryNum; i++) {
      if (m_pDOI[i] < 1) continue;

      m_pRenderColors[iCounter] = m_pSelectedColor;
      iCounter++;
    }

    //5. set the update flags
    m_bUpdateSelectionColoring = false;
    m_bUpdateSoftSelectionColoring = false;



  }


  private void recalculateSoftColoring() {

    //1. recalculate the color transitions
    if (m_pRenderColors == null) return;
    if (m_pDOI == null) return;
    int iCounter = m_iNotSelectedNum;
    int iEntryNum = m_pRenderColors.length;

    //1.1. set the colors of the soft selected items
    if (m_pSoftSelectionColorTransition != null) {
      //an advanced colortransition is there
      for (int i = 0; i < iEntryNum; i++) {
        if ((m_pDOI[i] > 0) && (m_pDOI[i] < 1)) {
          m_pRenderColors[iCounter] = m_pSoftSelectionColorTransition.getColorForPos(m_pDOI[i]);
          iCounter++;
        }
      }
    }
    else {
      //1.2. there is no advanced color transition

      //components of the selected color if a simple linear transition is used
      float fRed1 = (float)m_pSelectedColor.getRed();
      float fGreen1 = (float)m_pSelectedColor.getGreen();
      float fBlue1 = (float)m_pSelectedColor.getBlue();

      //components of the not selected color
      float fRed2 = (float)m_pNotSelectedColor.getRed();
      float fGreen2 = (float)m_pNotSelectedColor.getGreen();
      float fBlue2 = (float)m_pNotSelectedColor.getBlue();

      for (int i = 0; i < iEntryNum; i++) {
        float fDOI = m_pDOI[i];
        if ((fDOI > 0) && (fDOI < 1)) {
          float fInvDOI = 1 - fDOI;
          int iRed = (int)(fRed1*fDOI + fInvDOI*fRed2);
          int iGreen = (int)(fGreen1*fDOI + fInvDOI*fGreen2);
          int iBlue = (int)(fBlue1*fDOI + fInvDOI*fBlue2);

          m_pRenderColors[iCounter] = new Color(iRed, iGreen, iBlue);
          iCounter++;
        }
      }
    }

    //2. set the update flags
    m_bUpdateSoftSelectionColoring = false;


  }





}