package view;

/*
 *  Copyright (C) 2018 Juergen Platzer 
 * 
 *  This file is part of JInfoViz.
 *  
 *  JInfoViz is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU Lesser General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *  
 *  JInfoViz is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU Lesser General Public License for more details.
 *  
 *  You should have received a copy of the GNU Lesser General Public License
 *  along with JInfoViz.  If not, see <http://www.gnu.org/licenses/>.
 */


import listeners.ITransitionPanelListener;

import javax.swing.*;
import java.awt.*;
import java.awt.event.*;
import java.util.Vector;


/** class implementing the view of a color transition **/
public class TransitionPanel extends JPanel implements MouseListener, MouseMotionListener{

  /*** varables that steer the layout ***/
  private ColorTransition m_pLinearTransition, m_pAdvancedTransition;   //2 different color transitions (= 2 modes)
  private boolean m_bShowAdvancedTransition;                //flag indicating which mode (true = advanced mode)

  private String m_pLeftText, m_pCenterText, m_pRightText;   //Strings that should be written over the transition
  private boolean m_bShowText;                  //flag indicating if a text should be written.

  /** variables to navigate the steering color points **/
  private int m_iSteeringOffset, m_iSteeringHeight, m_iDotWidth;
  private ColorPosition m_pCurrentColorPosition;
  private boolean m_bDragPoint;               //indicates if a color dot is dragged

  //Vector holding all objects that listen to changes of this transition panel
  private Vector<ITransitionPanelListener> m_pTransitionPanelListeners; 

  
  //////////////////////////
  // CONSTRUCTORS         //
  //////////////////////////
  public TransitionPanel(Color pStartColor, Color pEndColor) {
    //1. make some checks
    if (pStartColor == null) pStartColor = Color.BLACK;
    if (pEndColor == null) pEndColor = Color.RED;
    
    //2. create a linear color transition
    ColorPosition[] pColorPositions = new ColorPosition[2];
    pColorPositions[0] = new ColorPosition(pStartColor, 0.0f);
    pColorPositions[1] = new ColorPosition(pEndColor, 1.0f);
    m_pLinearTransition = new ColorTransition(pColorPositions);
    
    //3. set the advanced color transition variables to the default
    m_pAdvancedTransition = null;
    m_bShowAdvancedTransition = false;

    //4. set the default text variable values
    m_pLeftText = null;
    m_pCenterText = null;
    m_pRightText = null;
    m_bShowText = false;
    
    //5. set the default interaction values
    m_pCurrentColorPosition = null;
    m_bDragPoint = false;


    //6. add listener functionality
    addMouseListener(this);
    addMouseMotionListener(this);

    

    //7. init the listener vector
    m_pTransitionPanelListeners = new Vector<ITransitionPanelListener>();



  }




  ////////////////////////////////////
  // PAINTING                       //
  ////////////////////////////////////


  //calls the paint-method
  public void paintComponent(Graphics g) {
    super.paintComponent(g);
    update(g);
  }

  //paint-method
  public void update(Graphics g) {

    //1. prepare the graphics context
    Graphics2D pGraphics = (Graphics2D)g;

    //2. clear background
    pGraphics.setColor(getBackground());
    int iWidth = getWidth();
    int iHeight = getHeight();
    
    pGraphics.fillRect(0,0, iWidth, iHeight);

    //3. vertical offset for the drawing of the transition bar
    int iVerticalOffset = 0;


    //4. write text if necessary
    if (m_bShowText) {

      //4.1. make some settings
      iVerticalOffset = 25; //offset is set -> text is not written into the transition bar
      pGraphics.setColor(Color.BLACK);

      //4.2. write the left text
      if (m_pLeftText != null) {
        pGraphics.drawString(m_pLeftText, 0, 20);
      }

      
      //4.3. write the center text
      FontMetrics pMetrics = pGraphics.getFontMetrics();
      if (m_pCenterText != null) {
        int iStringWidth = pMetrics.stringWidth(m_pCenterText);
        pGraphics.drawString(m_pCenterText, (iWidth-iStringWidth)/2, 20);
      }

      //4.4. write the right text
      if (m_pRightText != null) {
        int iStringWidth = pMetrics.stringWidth( m_pRightText );
        pGraphics.drawString(m_pRightText, iWidth-iStringWidth, 20);
      }

    }


    //5. advanced transition, if set
    if (m_bShowAdvancedTransition) {

      //DRAW COLOR TRANSITION
      //5.1. draw start color rectangle ( the start color from 0 to first color position)
      int iTransitionHeight = (iHeight - iVerticalOffset) / 2;
      
      ColorPosition pStartColorPosition = m_pAdvancedTransition.getColorPosition(0);
      pGraphics.setColor(pStartColorPosition.getColor());
      pGraphics.fillRect(0, iVerticalOffset, (int)(pStartColorPosition.getPosition()*iWidth), iTransitionHeight);

      //5.2. draw transition
      int iColorPosNum = m_pAdvancedTransition.getColorPositionNumber();
      
      for (int i = 0; i < iColorPosNum-1; i++) {
        //fetch the color positions of interest
        ColorPosition pColorPosition0 = m_pAdvancedTransition.getColorPosition(i);
        ColorPosition pColorPosition1 = m_pAdvancedTransition.getColorPosition(i+1);

        //create the gradient paint for filling the rectangle
        float fHorizontalOffset = pColorPosition0.getPosition()*iWidth;
        float fPositionGap = pColorPosition1.getPosition()*iWidth - fHorizontalOffset + 1;
        GradientPaint pGradientPaint = new GradientPaint(fHorizontalOffset, 0.0f, pColorPosition0.getColor(),
        fHorizontalOffset + fPositionGap, 0.0f, pColorPosition1.getColor());

        //fill the rectangle
        pGraphics.setPaint(pGradientPaint);
        pGraphics.fillRect((int)fHorizontalOffset, iVerticalOffset, (int)fPositionGap, iTransitionHeight);
      }

      //5.3. draw end color rectangle ( the end color from last color position to width)
      ColorPosition pEndColorPosition = m_pAdvancedTransition.getColorPosition(iColorPosNum-1);
      pGraphics.setColor(pEndColorPosition.getColor());
      float fHorizontalOffset = pEndColorPosition.getPosition()*iWidth;
      float fPositionGap = iWidth - fHorizontalOffset;
      pGraphics.fillRect((int)fHorizontalOffset, iVerticalOffset, (int)fPositionGap, iTransitionHeight);


      //DRAW STEERING BAR
      //5.4. draw steering color dots
      m_iSteeringOffset = iVerticalOffset + iTransitionHeight;
      m_iSteeringHeight = iHeight - m_iSteeringOffset;
      m_iDotWidth = m_iSteeringHeight / 2;

      for (int i = 0; i < iColorPosNum; i++) {
        ColorPosition pColorPosition = m_pAdvancedTransition.getColorPosition(i);
        int iPosition = (int)(iWidth*pColorPosition.getPosition());

        //the color dot
        pGraphics.setColor(pColorPosition.getColor());
        pGraphics.fillRect(iPosition - m_iDotWidth/2, m_iSteeringOffset + m_iDotWidth/2, m_iDotWidth, m_iDotWidth);

        //the black color frame
        pGraphics.setColor(Color.BLACK);
        pGraphics.drawRect(iPosition - m_iDotWidth/2, m_iSteeringOffset + m_iDotWidth/2, m_iDotWidth, m_iDotWidth);

      }

      //5.5. draw black frame
      pGraphics.setColor(Color.BLACK);
      pGraphics.drawRect(0, m_iSteeringOffset, iWidth-1, m_iSteeringHeight-1);




    }

    //6. linear transition
    else {


      //Draw color transition
      ColorPosition pColorPos0 = m_pLinearTransition.getColorPosition(0);
      ColorPosition pColorPos1 = m_pLinearTransition.getColorPosition(1);

      GradientPaint pGradientPaint = new GradientPaint(pColorPos0.getPosition()*iWidth, 0.0f, pColorPos0.getColor(),
      pColorPos1.getPosition()*iWidth, 0.0f, pColorPos1.getColor());

      pGraphics.setPaint(pGradientPaint);
      pGraphics.fillRect(0,iVerticalOffset, iWidth, iHeight-iVerticalOffset);

    }
  }



  ///////////////////////////////////
  // SET METHODS                   //
  ///////////////////////////////////
  /** sets a new linear transition**/
  public void setLinearTransition(Color pStartColor, Color pEndColor) {
    //1. make some checks
    if (pStartColor == null) return;
    if (pEndColor == null) return;
  
    //2. create a new linear transition
    ColorPosition[] pColorPositions = new ColorPosition[2];
    pColorPositions[0] = new ColorPosition(pStartColor, 0.0f);
    pColorPositions[1] = new ColorPosition(pEndColor, 1.0f);
    m_pLinearTransition = new ColorTransition(pColorPositions);
    
    repaint();

  }

  /** sets the advanced transition or the linear transition visible **/
  public void setAdvancedTransitionVisible(boolean bShowAdvancedTransition) {
    if ((bShowAdvancedTransition) && (m_pAdvancedTransition == null)) {
      m_pAdvancedTransition = new ColorTransition(m_pLinearTransition);
    }
    m_bShowAdvancedTransition = bShowAdvancedTransition;
    repaint();

  }

  /** the text that should be written on the transition **/
  public void setText(String pLeftText, String pCenterText, String pRightText) {
    
    //1. set the text strings
    m_pLeftText = pLeftText;
    m_pCenterText = pCenterText;
    m_pRightText = pRightText;

    //2. calculate the show text flag
    m_bShowText = (m_pLeftText != null) || (m_pCenterText != null) || (m_pRightText != null);

    //3. init a repaint
    repaint();
  }







  ////////////////////////////////////
  // MOUSE LISTENER                 //
  ////////////////////////////////////

  /** open color chooser with left mouse click **/
  public void mouseClicked(MouseEvent e) {

    //1. make some checks
    if (!m_bShowAdvancedTransition) return;
    if (m_pAdvancedTransition == null) return;

    //2. handle a left mouse click -> choose a new color
    if (e.getButton() == MouseEvent.BUTTON3) {
      //check if mouseclick is on a steering colordot
      if (isOnSteeringColorDot(e.getX(), e.getY())) {
        Color pNewColor = JColorChooser.showDialog(this, "Choose new color", m_pCurrentColorPosition.getColor());
        if (pNewColor != null) m_pCurrentColorPosition.setColor(pNewColor);

        m_pCurrentColorPosition = null;
        repaint();
        transitionPanelChanged();
        return;
      }
    }

    //3. double click -> delete the current color dot or add a new one
    if (e.getClickCount() == 2) {
      //delete the current colordot, when the double click is on a color dot
      if (isOnSteeringColorDot(e.getX(), e.getY())) {
        m_pAdvancedTransition.deleteColorPosition(m_pCurrentColorPosition);
        m_pCurrentColorPosition = null;
        repaint();
        transitionPanelChanged();
        return;
      }

      //add a new colorposition
      float fPosition = (float)e.getX()/(float)getWidth();
      Color pColor = m_pAdvancedTransition.getColorForPos(fPosition);
      ColorPosition pColorPosition = new ColorPosition(pColor, fPosition);
      m_pAdvancedTransition.addColorPosition(pColorPosition);

      repaint();
      transitionPanelChanged();
      return;
    }


  }

  public void mouseEntered(MouseEvent e) {}
  public void mouseExited(MouseEvent e)  {}


  //start of dragging a color dot
  public void mousePressed(MouseEvent e)  {
    //1. make some checks -> only for advanced transition and left mouse button
    if (!m_bShowAdvancedTransition) return;
    if (m_pAdvancedTransition == null) return;
    if (e.getButton() != MouseEvent.BUTTON1) return;

    //2. set the drag point flag, if the action is on a color dot
    if (isOnSteeringColorDot(e.getX(), e.getY())) {
      m_bDragPoint = true;
    }
  }



  //end of dragging a color dot
  public void mouseReleased(MouseEvent e) {
    //1. make a check: if no point has been dragged -> nothing is to do
    if (!m_bDragPoint) return;

    //2. if a dragging operation has been made, just do the reposition of the color point...
    mouseDragged(e);
    
    //...  and set the dragging variables to the defaults
    m_bDragPoint = false;
    m_pCurrentColorPosition = null;
    transitionPanelChanged();
  }



  ////////////////////////////////////
  // MOUSE MOTION LISTENER          //
  ////////////////////////////////////
  //dragging a color dot
  public void mouseDragged(MouseEvent e) {
    if (m_bDragPoint) {
      float fPosition = (float)e.getX()/(float)getWidth();
      m_pCurrentColorPosition.setPosition(fPosition);
      repaint();
    }
  }


  public void mouseMoved(MouseEvent e) {}



  ////////////////////////////////////
  // CALCULATIONS FOR THE MOUSE     //
  // INTERACTION                    //
  ////////////////////////////////////
  /** returns true if a coordinate is on one colordot and sets the current ColorPosition */
  private boolean isOnSteeringColorDot(int iXPos, int iYPos) {
    //1. check y position
    if (iYPos < m_iSteeringOffset + m_iDotWidth/2) return false;
    if (iYPos > m_iSteeringOffset + 3*m_iDotWidth/2) return false;

    //2. check x position and return true and set the current color point if the coordinates are on a color point. 
    int iColorPosNum = m_pAdvancedTransition.getColorPositionNumber();
    for (int i = 0; i < iColorPosNum; i++) {
      ColorPosition pColorPosition = m_pAdvancedTransition.getColorPosition(i);
      int iPosition = (int)(getWidth()*pColorPosition.getPosition());

      if ((iXPos >= iPosition - m_iDotWidth/2) && (iXPos <= iPosition + m_iDotWidth/2)) {
        m_pCurrentColorPosition = pColorPosition;
        return true;
      }
    }

    return false;
  }

  ///////////////////////////////////
  // GET METHODS                   //
  ///////////////////////////////////
  public ColorTransition getAdvancedTransition() {
    if (m_bShowAdvancedTransition) return m_pAdvancedTransition;

    return null;
  }



  //////////////////////////////////////////
  // TRANSITIONPANEL LISTENER METHODS     //
  //////////////////////////////////////////
  public void addTransitionPanelListener(ITransitionPanelListener pListener) {
    if (pListener == null) return;
    m_pTransitionPanelListeners.add(pListener);
  }

  private void transitionPanelChanged() {
    int iListenerNum = m_pTransitionPanelListeners.size();
    for (int i = 0; i < iListenerNum; i++) {
      ITransitionPanelListener pListener = m_pTransitionPanelListeners.get(i);
      pListener.transitionPanelChanged(this);
    }

  }



}